/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_BIT_COUNT_VISITOR__
#define __DD_BIT_COUNT_VISITOR__

#include <vector>
#include <algorithm>

#include "DDBitList.h"
#include "DDRayWrapperVisitor.h"

namespace addibit {

    class BitPopulationEntry {

        private:
        unsigned int mPosition;
        unsigned long mCount;

        public:
        BitPopulationEntry( const unsigned int pos ) :
            mPosition( pos ),
            mCount( 0 ) {}

        bool operator< ( const BitPopulationEntry & rhs ) const {
            return mCount > rhs.mCount;
        }

        void reset ( unsigned int pos ) {
            mCount = 0;
            mPosition = pos;
        }

        void inc () {
            mCount++;
        }

        unsigned int position () const {
            return mPosition;
        }

        unsigned long count () const {
            return mCount;
        }
    };

    template<class RAY_TYPE>
    class BitCountVisitor : public RayWrapperVisitor<RAY_TYPE> {

        private:
        uintptr_t mKey;
        std::vector<unsigned long> mStats;
        uint_fast16_t mCapacity;

        uint32_t * __stats;
        uint_fast16_t __stats_size;

        std::vector<BitPopulationEntry> mPopulation;
        unsigned long mTotalSetBits;

        const __m128i __lookup_table[16] = {
            _mm_set_epi32(0, 0, 0, 0),
            _mm_set_epi32(0, 0, 0, 1),
            _mm_set_epi32(0, 0, 1, 0),
            _mm_set_epi32(0, 0, 1, 1),
            _mm_set_epi32(0, 1, 0, 0),
            _mm_set_epi32(0, 1, 0, 1),
            _mm_set_epi32(0, 1, 1, 0),
            _mm_set_epi32(0, 1, 1, 1),
            _mm_set_epi32(1, 0, 0, 0),
            _mm_set_epi32(1, 0, 0, 1),
            _mm_set_epi32(1, 0, 1, 0),
            _mm_set_epi32(1, 0, 1, 1),
            _mm_set_epi32(1, 1, 0, 0),
            _mm_set_epi32(1, 1, 0, 1),
            _mm_set_epi32(1, 1, 1, 0),
            _mm_set_epi32(1, 1, 1, 1)
        };

        void get_stats ( __BIT_LIST_RAW_DATA_TYPE__ data, const uint_fast16_t offset ) {

            __m128i * __mem_addr = reinterpret_cast<__m128i *>( __stats + offset );
            const __m128i * __end_addr = reinterpret_cast<const __m128i *>( __stats + (( __stats_size - offset < 64 ) ? __stats_size : ( offset + 64 ) ));
            
            do {
                _mm_store_si128( __mem_addr, _mm_add_epi32( _mm_load_si128( __mem_addr ), __lookup_table[ data & 0x000000000000000Ful ] ));
                data >>= 4;
            } while ( ++__mem_addr != __end_addr );
        }

        public:
        BitCountVisitor() :
            mKey( 0 ),
            mCapacity( 0 ),
            __stats( NULL ) {}

        virtual ~BitCountVisitor () {
            if ( __stats ) {
                free( __stats );
            }
        }

        void set_key ( uintptr_t key ) {
            mKey = key;
        }
        
        void set_capacity ( const uint_fast16_t capacity ) {
            mCapacity = capacity;
        }
        
        uintptr_t key () const {
            return mKey;
        }

        void onStart ( const uint_fast32_t );
        void process_ray ( const RAY_TYPE & );
        void onFinish ();

        const std::vector<unsigned long> stats () const {
            return mStats;
        }

        const std::vector<BitPopulationEntry> & population () const {
            return mPopulation;
        }

        double getTotalSetBits () const {
            return mTotalSetBits;
        }

        const BitList getMostCommonBits ( unsigned int ) const;
        const BitList * getMostCommonBitsRange ( const unsigned int, const unsigned int );
        const BitList getFirstNMostCommonBits ( const unsigned int, const unsigned int ) const;
        
        const BitList getFirstNMostCommonBits ( const unsigned int n ) const {
            return getFirstNMostCommonBits( n, 0 );
        }
    };
    
    template<class RAY_TYPE>
    void BitCountVisitor<RAY_TYPE>::onStart ( const uint_fast32_t ) {
        mStats.assign( mCapacity, 0 );
        __stats_size = mCapacity + (( mCapacity % 4 ) ? ( 4 - ( mCapacity % 4 )) : 0 );

        __stats = new ( aligned_alloc( 16, __stats_size * sizeof( uint32_t ) ) ) uint32_t[ __stats_size ];

        for ( uint_fast16_t pos = 0; pos < __stats_size; pos++ ) {
            __stats[ pos ] = 0;
        }     
        mTotalSetBits = 0;      
    }

    template<class RAY_TYPE>
    void BitCountVisitor<RAY_TYPE>::process_ray ( const RAY_TYPE & ray ) {
        uint_fast16_t i = 0;        
        const __BIT_LIST_RAW_DATA_TYPE__ * data = ray.getActiveSet().data();
        
        do {
            get_stats( data[i], i * 64 );
        } while ( ++i != ( __stats_size + 63 ) / 64 );
    }

    template<class RAY_TYPE>
    void BitCountVisitor<RAY_TYPE>::onFinish () {

        for ( uint_fast16_t pos = 0; pos < mCapacity; pos++ ) {
            mStats[ pos ] = __stats[ pos ];
            mTotalSetBits += __stats[ pos ];
        } 
    }

    template<class RAY_TYPE>
    const BitList BitCountVisitor<RAY_TYPE>::getMostCommonBits ( unsigned int percentage ) const {
        BitList result( __BIT_LIST_ALLOC_SPACE__ );
        unsigned long current_set_bits = 0;

        for ( unsigned long i = 0; i < mPopulation.size() && ( 100.0 * current_set_bits ) / mTotalSetBits < percentage; i++ ) {
            result.setBit( mPopulation[ i ].position() );
            current_set_bits += mPopulation[ i ].count();
        }

        return result;
    }

    template<class RAY_TYPE>
    const BitList * BitCountVisitor<RAY_TYPE>::getMostCommonBitsRange ( const unsigned int n, const unsigned int range ) {
        BitList * result = new BitList[ range ];
        
        result[ 0 ] = getFirstNMostCommonBits( n );

        for ( unsigned int i = 1; i < range; i++ ) {
            result[ i ] = result[ i - 1 ];
            result[ i ].setBit( mPopulation[ n + 1 ].position() );
        }

        return result;
    }

    template<class RAY_TYPE>
    const BitList BitCountVisitor<RAY_TYPE>::getFirstNMostCommonBits ( const unsigned int n, const unsigned int start ) const {
        BitList result( __BIT_LIST_ALLOC_SPACE__ );

        for ( unsigned long i = start; i < mPopulation.size() && i < (start + n); i++ ) {
            result.setBit( mPopulation[ i ].position() );
        }

        return result;
    }    
}

#endif // __DD_BIT_COUNT_VISITOR__
