/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_BIT_LIST__
#define __DD_BIT_LIST__

#include <vector>
#include <string>
#include <string.h>

#include "DDTweaks.h"
#include "DDConstants.h"
#include "DDLogging.h"

#include "DDBitCountTable.h"

#if defined _LP64 || defined __LP64__
#define __SINGLE_BIT__ ( static_cast<unsigned long>( 0x0000000000000001 ))
#define __FULL_BLOCK__ 0xFFFFFFFFFFFFFFFF
#define __BLOCK_SIZE__ 64
#else 
#define __SINGLE_BIT__ 0x00000001
#define __FULL_BLOCK__ 0xFFFFFFFF
#define __BLOCK_SIZE__ 32
#endif // defined _LP64 || defined __LP64__


#define __BIT_LIST_ALLOC_SPACE__  new unsigned long[ __DD_BLOCK_COUNT__ ]
#define __BIT_LIST_RAW_DATA_TYPE__ unsigned long

namespace addibit {

    class BitListData {
        
        private:
        __BIT_LIST_RAW_DATA_TYPE__ data[ 1 ];
        
        public:
        
        inline __BIT_LIST_RAW_DATA_TYPE__ operator[] ( const uint_fast16_t index ) const {
            return data[ index ];
        }
        
        inline __BIT_LIST_RAW_DATA_TYPE__ & operator[] ( const uint_fast16_t index ) {
            return data[ index ];
        }
    };

	class BitList {

		private:
		__BIT_LIST_RAW_DATA_TYPE__ * __data;  

		public:
        static uint_fast16_t LAST_POP_CNT_RESULT;
        
        BitList ( __BIT_LIST_RAW_DATA_TYPE__ * _data = NULL ) :
            __data( _data ) {}

		BitList ( __BIT_LIST_RAW_DATA_TYPE__ * _data, const unsigned long value ) :	
            __data( _data ) {
        
			__data[ 0 ] = value;

			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__; block++ ) {
				__data[ block ] = value;
			}             
        } 

        BitList ( const BitList & lvalue ) :
            __data( __BIT_LIST_ALLOC_SPACE__ ) {

			__data[ 0 ] = lvalue.__data[ 0 ];

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; ++block ) {
				__data[ block ] = lvalue.__data[ block ];
			} 
		}
        
        BitList ( BitList && rvalue ) :
            __data( rvalue.__data ) {
                
            rvalue.reset();
        }
        
        /* Used for creation of new rays */
        BitList ( BitList && edge_set,
                  const uint_fast16_t p ) :
            __data( edge_set.__data ) {
                
            edge_set.reset();
            setBit( p );
        }

		~BitList () {
            if ( unlikely( __data != NULL ) ) {
                delete[] __data;
            }
		}

		static const BitCountTable BIT_COUNT_TABLE;

		static uint_fast16_t getBlocks ( const uint_fast16_t capacity ) {
			return capacity / __BLOCK_SIZE__ + (( capacity % __BLOCK_SIZE__ == 0 ) ? 0 : 1 );
		}

		unsigned long getCapacity () const {
			return __DD_BLOCK_COUNT__ * __BLOCK_SIZE__;
		}

		void init ( const unsigned int blocks, const unsigned long data );

		void resize ( const unsigned long blocks );

		bool isEmpty () const {
			bool result = ( getData( 0 ) == 0 );

			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__ && result; block++ ) {
				if ( getData( block ) != 0 ) {
					return false;
				}
			}

			return result;
		}

		bool isFull () const {
			bool result = ( getData( 0 ) == ULONG_MAX );

			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__ && result; block++ ) {
				if ( getData( block ) != ULONG_MAX ) {
					result = false;
					break;
				}
			}

			return result;
		}

		bool isUnionFull ( const BitList & other ) const {
			bool result = ( (getData( 0 ) | other.getData( 0 )) == ULONG_MAX );

			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__ && result; block++ ) {
				if ( (getData( block ) | other.getData( block )) != ULONG_MAX ) {
					result = false;
					break;
				}
			}

			return result;
		}

        inline const __BIT_LIST_RAW_DATA_TYPE__ & getData ( const uint_fast16_t block ) const {	
			return __data[ block ];
		}

		inline __BIT_LIST_RAW_DATA_TYPE__ & getData ( const uint_fast16_t block ) {
			return __data[ block ];
		}
        
        inline const __BIT_LIST_RAW_DATA_TYPE__ * data () const {
            return __data;
        }

        BitList operator & ( const BitList & rhs ) const {
			BitList result( __BIT_LIST_ALLOC_SPACE__ );

			for ( uint_fast16_t i = 0; i < __DD_BLOCK_COUNT__; i++ ) {
				result[i] = getData( i ) & rhs[i];
			}

			return result;
		}
        
        static uint_fast16_t _op_cut_and_save ( __BIT_LIST_RAW_DATA_TYPE__ * const dst,
                                                const __BIT_LIST_RAW_DATA_TYPE__ * const arg_1,
                                                const __BIT_LIST_RAW_DATA_TYPE__ * const arg_2 ) {
                                           
            dst[0] = arg_1[0] & arg_2[0];
            uint_fast16_t popcount = BitList::BIT_COUNT_TABLE.lookUp ( dst[0] );
            
			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__; ++block ) {
				dst[ block ] = arg_1[ block ] & arg_2[ block ];
                popcount +=  BitList::BIT_COUNT_TABLE.lookUp ( dst[ block ] );
			}   
            
            return popcount;   
        }
        
   /*     inline BitList & _op_cut_and_save ( const __BIT_LIST_RAW_DATA_TYPE__ * const arg_1, const __BIT_LIST_RAW_DATA_TYPE__ * const arg_2 ) {
            _op_cut_and_save( __data, arg_1, arg_2 );
            return *this;
        } */

		const BitList operator| ( const BitList& rhs ) const {
			BitList result( __BIT_LIST_ALLOC_SPACE__ );

			for ( uint_fast16_t i = 0; i < __DD_BLOCK_COUNT__; i++ ) {
				result[i] = getData( i ) | rhs[i];
			}

			return result;
		}

		const BitList operator^ ( const BitList & rhs ) const {
			BitList result( __BIT_LIST_ALLOC_SPACE__ );

            result[ 0 ] = __data[ 0 ] ^ rhs[ 0 ];

			for ( uint_fast16_t i = 1; i < __DD_BLOCK_COUNT__; i++ ) {
				result[ i ] = getData( i ) ^ rhs.getData( i );
			}

			return result;
		}

		const BitList operator~ () const {
			BitList result( __BIT_LIST_ALLOC_SPACE__ );

			for ( uint_fast16_t i = 0; i < __DD_BLOCK_COUNT__; i++ ) {
				result[i] = ~getData( i );
			}

			return result;
		}

        const BitList operator- ( const BitList & rhs ) const {
            return (*this) ^ ( (*this) & rhs );
        }

		bool operator== ( const BitList & rhs ) const {

			for ( uint_fast16_t i = 0; i < __DD_BLOCK_COUNT__; i++ ) {
				if ( getData( i ) != rhs[i] ) {
					return false;
			    }
            } 

			return true;
		}

		bool operator!= ( const BitList & rhs ) const {
	
			return !operator==( rhs );
		}

        bool lex_less ( const BitList & rhs ) const {
            uint_fast16_t block = 0;                  
      
            while ( block + 1 < __DD_BLOCK_COUNT__ && getData( block ) == rhs.getData( block ) ) {
                ++block;
            }

            return getData( block ) < rhs.getData( block );
        }
        
		void operator ^= ( const BitList& rhs ) {
			__data[0] ^= rhs[0];			

			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__; block++ ) {
				__data[ block ] ^= rhs[ block ];
			}
		}    
		
		/* Precondition: shift < 64! */
		void operator <<= ( const unsigned int shift ) {

            unsigned long transfer = __data[ 0 ] >> ( __BLOCK_SIZE__ - shift );
			__data[ 0 ] <<= shift;			

			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__; block++ ) {
			    const unsigned long buffer = __data[ block ] >> ( __BLOCK_SIZE__ - shift );
				__data[ block ] = ( __data[ block ] << shift ) | transfer;
				transfer = buffer;
			}
		}     		    

		void operator &= ( const BitList& rhs ) {

			__data[0] &= rhs[0];			

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				__data[block] &= rhs[block];
			}
		}

		void operator |= ( const BitList & rhs ) {

			__data[0] |= rhs[0];			

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				__data[block] |= rhs[block];
			}
		}

		void operator= ( const BitList & lvalue ) {
            uint_fast16_t i = 0;
            
            do {
                __data[i] = lvalue[i];			
            } while ( ++i < __DD_BLOCK_COUNT__ );
		}
        
        BitList & operator= ( BitList && rvalue ) {
            std::swap( __data, rvalue.__data );
            return *this;
        }

        inline static void copy ( const __BIT_LIST_RAW_DATA_TYPE__ * const src, __BIT_LIST_RAW_DATA_TYPE__ * const dst ) {
            dst[0] = src[0];
            
			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				dst[ block ] = src[ block ];
			}            
        }

		const unsigned long& operator[] ( const unsigned long index ) const {
			return getData( index );
		}
		unsigned long& operator[] ( const unsigned long index ) {
			return getData( index );
		}

		void clear () {

			__data[0] = 0;			

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				__data[block] = 0;
			}
		}

        static uint_fast16_t getPopulationCount ( const __BIT_LIST_RAW_DATA_TYPE__ * const __map ) {
			uint_fast16_t count = BitList::BIT_COUNT_TABLE.lookUp ( __map[0] );			

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( __map[ block ] );
			}

			return count;
        }

		inline uint_fast16_t getPopulationCount () const {
			return getPopulationCount( __data );
		}

		uint_fast16_t getInvPopulationCount () const {
			uint_fast16_t count = BitList::BIT_COUNT_TABLE.lookUp ( ~__data[0] );			

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( ~getData( block ) );
			}

			return count;
		}

        uint_fast16_t getHammingDistance ( const BitList & other ) const {
			uint_fast16_t count = BitList::BIT_COUNT_TABLE.lookUp ( __data[0] xor other.__data[0] );			

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) xor other.getData( block ) );
			}

			return count;
        }
        
        static uint_fast16_t getReducedCommonPopCnt ( const __BIT_LIST_RAW_DATA_TYPE__ * const node_data,
                                                      const __BIT_LIST_RAW_DATA_TYPE__ * const ref_data,
                                                      const __BIT_LIST_RAW_DATA_TYPE__ * const reduction ) {
            
			uint_fast16_t count = BitList::BIT_COUNT_TABLE.lookUp ( node_data[0] & ((( ref_data[0] & reduction[0] ) >> 1 ) | ref_data[0] ) );			

			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__; ++block ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( node_data[block] & ((( ref_data[block] & reduction[block] ) >> 1 ) | ref_data[block] ));
			}

			return count;            
        }

		inline uint_fast16_t getReducedCommonPopCnt ( const __BIT_LIST_RAW_DATA_TYPE__ * other, const BitList & reduction ) const {
			return getReducedCommonPopCnt( __data, other, reduction.data() );
		}

		inline uint_fast16_t getReducedCommonPopCnt ( const BitList & other, const BitList & reduction ) const {
			return getReducedCommonPopCnt( other.data(), reduction );
		}

        static uint_fast16_t getCommonPopCnt ( const __BIT_LIST_RAW_DATA_TYPE__ * const arg_1, const __BIT_LIST_RAW_DATA_TYPE__ * const arg_2 ) {
			uint_fast16_t count = BitList::BIT_COUNT_TABLE.lookUp ( arg_1[0] & arg_2[0] );			

			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( arg_1[ block ] & arg_2[ block ] );
			}

			return count;            
        }

        inline uint_fast16_t getCommonPopCnt ( const __BIT_LIST_RAW_DATA_TYPE__ * const other_data ) const {
			return getCommonPopCnt( __data, other_data );
		}

        inline uint_fast16_t getCommonPopCnt ( const BitList & other ) const {
			return getCommonPopCnt( other.__data );
		}
        
		inline void writeCommonPopCnt ( unsigned long * & lvalue ) const {
			LAST_POP_CNT_RESULT = BitList::BIT_COUNT_TABLE.lookUp ( __data[0] & ( lvalue++ )[ 0 ] );			

			for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__; block++ ) {
				LAST_POP_CNT_RESULT += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) & ( lvalue++ )[ 0 ] );
			}
		}
        
        static void writeCommonAndDiffPopCnt ( const __BIT_LIST_RAW_DATA_TYPE__ * const __a,
                                               const __BIT_LIST_RAW_DATA_TYPE__ * const __b,
                                               uint_fast16_t result[2] ) {
            
            /* Common Pop Cnt*/
            result[ 0 ] = BitList::BIT_COUNT_TABLE.lookUp( __a[0] & __b[0] );
            
            /* Diff Pop Cnt */
            result[ 1 ] = BitList::BIT_COUNT_TABLE.lookUp( __a[0] & __b[1] );
        }
        
        
        static uint_fast16_t pop_cnt_a_minus_b ( const __BIT_LIST_RAW_DATA_TYPE__ * const __a, const __BIT_LIST_RAW_DATA_TYPE__ * const __b ) {
			uint_fast16_t count = 0;
            uint_fast16_t i = 0;

            do {
                count += __builtin_popcount( ( __a[i] ^ __b[i] ) & __a[i] );
            } while ( ++i < __DD_BLOCK_COUNT__ );

			return count;
		}

        static __m64 nar_pop_cnt_a_minus_b ( const __BIT_LIST_RAW_DATA_TYPE__ * const __a, const __BIT_LIST_RAW_DATA_TYPE__ * const __b ) {
            uint_fast16_t i = 0;
            __m64 count = _mm_setzero_si64();

            do {
                count = _mm_add_pi32( count, _mm_set_pi32( _mm_popcnt_u64( __a[i] & __b[i] ), _mm_popcnt_u64( ( __a[ i + __DD_BLOCK_COUNT__ ] ^ __b[i] ) & __a[ i + __DD_BLOCK_COUNT__ ] ) ));
            } while ( ++i < __DD_BLOCK_COUNT__ );

			return count;
		}
        
        static bool ver_pop_cnt_a_minus_b ( const __BIT_LIST_RAW_DATA_TYPE__ * const __a, const __BIT_LIST_RAW_DATA_TYPE__ * const __b, const uint_fast16_t inf ) {
			uint_fast16_t count = 0;
            uint_fast16_t i = 0;

            do {
                count += ( ~__a[i] & __b[i] ) ? inf : __builtin_popcount( ( __a[ i + __DD_BLOCK_COUNT__ ] ^ __b[i] ) & __a[ i + __DD_BLOCK_COUNT__ ] );
            } while ( count < inf && ++i < __DD_BLOCK_COUNT__ );

			return i == __DD_BLOCK_COUNT__;
		}
        
		uint_fast16_t getDiffPopCnt ( const __BIT_LIST_RAW_DATA_TYPE__ * const other ) const {
			uint_fast16_t count = BitList::BIT_COUNT_TABLE.lookUp ( ( __data[0] ^ other[0] ) & __data[0] );			

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( ( getData( block ) ^ other[ block ] ) & getData( block ) );
			}

			return count;
		}	
        
        inline uint_fast16_t getDiffPopCnt ( const BitList & other ) const {
            return getDiffPopCnt( other.data() );
        }

        __m64 static getCommonAndInvPopCnt ( const __BIT_LIST_RAW_DATA_TYPE__ * const __node,  const __BIT_LIST_RAW_DATA_TYPE__ * const __ref ) {
            __m64 __result = _mm_setzero_si64();
            uint_fast16_t i = 0;
            
            do {
                _mm_add_pi32( __result, _mm_set_pi32( __builtin_popcount( ~__node[i] & __ref[i] ), __builtin_popcount( __node[i] & __ref[i] ) ));
            } while ( ++i < __DD_BLOCK_COUNT__ );
            
            return __result;
        }

		uint_fast16_t static getCommonPopCnt ( const __BIT_LIST_RAW_DATA_TYPE__ * const arg_1,  const __BIT_LIST_RAW_DATA_TYPE__ * const arg_2, uint_fast16_t & inv_pop_cnt ) {
            
			uint_fast16_t count = BitList::BIT_COUNT_TABLE.lookUp ( arg_1[0] & arg_2[0] );	
            inv_pop_cnt = BitList::BIT_COUNT_TABLE.lookUp ( ~arg_1[0] & arg_2[0] );		

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( arg_1[ block ] & arg_2[ block ] );
                inv_pop_cnt += BitList::BIT_COUNT_TABLE.lookUp ( ~arg_1[ block ] & arg_2[ block ] );
			}

			return count;
		}

		inline uint_fast16_t getCommonPopCnt ( const __BIT_LIST_RAW_DATA_TYPE__ * const other, uint_fast16_t & inv_pop_cnt ) const {
			return getCommonPopCnt( __data, other, inv_pop_cnt );
		}

        inline uint_fast16_t getCommonPopCnt ( const BitList & other, uint_fast16_t & inv_pop_cnt ) const {
            return getCommonPopCnt( other.data(), inv_pop_cnt );
        }

		uint_fast16_t getInvCommonPopCnt ( const BitList & other ) const {
			uint_fast16_t count = BitList::BIT_COUNT_TABLE.lookUp ( ~__data[0] & other.__data[0] );			

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( ~getData( block ) & other.getData( block ) );
			}

			return count;
		}

		uint_fast16_t getFirstSet () const {
			uint_fast16_t block = 0;

			for ( block = 0; unlikely( block + 1 < __DD_BLOCK_COUNT__ && !getData(block) ); block++ );

			return block * __BLOCK_SIZE__ + __builtin_ctzl( getData( block ));
		}
        
        inline uint_fast16_t getLastSet () const {
            uint_fast16_t block = __DD_BLOCK_COUNT__;
            uint_fast16_t pos;
            
            do {
                pos = __builtin_ffsl( getData( --block ) );
            } while ( unlikely( pos == 0 ) && block != 0 );
            
            return block * __BLOCK_SIZE__ + pos;
        }

		uint_fast16_t getFirstUnset () const {
			uint_fast16_t block = 0;

			for ( block = 0; unlikely( block + 1 < __DD_BLOCK_COUNT__ && !(~getData( block ))); block++ );

			return block * __BLOCK_SIZE__ + __builtin_ctzl( ~getData( block ));
		}

        uint_fast16_t getNextSet ( unsigned int start ) const {
            uint_fast16_t block = start / __BLOCK_SIZE__;
            uint_fast16_t offset = start % __BLOCK_SIZE__;
            uint_fast16_t result;

            unsigned long pattern = ( ULONG_MAX << offset ) & getData( block );

            if ( __builtin_expect( pattern != 0, 1 ) ) {
                result = block * __BLOCK_SIZE__ + __builtin_ctzl( pattern );
            } else if ( __builtin_expect( block + 1 >= __DD_BLOCK_COUNT__, 0 ) ) {
                result = getCapacity();
            } else {
                for ( block++ ; unlikely( block + 1 < __DD_BLOCK_COUNT__ && getData( block ) != 0 ); block++ );

                if ( getData( block ) != 0 ) {
                    result = block * __BLOCK_SIZE__ + __builtin_ctzl( getData( block ));
                } else {
                    result = getCapacity();
                }
            }

            return result;
        }

        static bool isSet ( const __BIT_LIST_RAW_DATA_TYPE__ * const __map, const uint_fast16_t position ) {
            return ( __map[ position / __BLOCK_SIZE__ ] & ( __SINGLE_BIT__ << ( position % __BLOCK_SIZE__ ) ) );
        }

		inline bool isSet ( const uint_fast16_t position ) const {
			return isSet( __data, position );
		}

		void setBit ( const uint_fast16_t position ) {
			getData( position / __BLOCK_SIZE__ ) |= ( __SINGLE_BIT__ << ( position % __BLOCK_SIZE__ ) );
		}

		void opOR ( const uint_fast16_t block, const unsigned long mask ) {
			getData( block ) |= mask;
		}

		void setBits ( const uint_fast16_t begin, const uint_fast16_t end ) {

			for ( uint_fast16_t index = begin; index < end; index++ ) {
				setBit( index );
			}
		}

        void set ( const unsigned long pattern ) {

            __data[ 0 ] = pattern;

            for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__; block++ ) {
			    __data[ block ] = pattern;
            }        
        }

		void unsetBits ( const uint_fast16_t begin, const uint_fast16_t end ) {
			for ( uint_fast16_t index = begin; index < end; index++ ) {
				unsetBit( index );
			}			
		}

		void unsetBitsFrom ( const uint_fast16_t begin ) {
			for ( uint_fast16_t index = begin; index < getCapacity() ; index++ ) {
				unsetBit( index );
			}			
		}

		void unsetBit ( const uint_fast16_t position ) {	
			getData( position/ __BLOCK_SIZE__ ) &= ~( __SINGLE_BIT__ << ( position % __BLOCK_SIZE__ ) );
		}

		void opXOR ( const uint_fast16_t block, const unsigned long mask ) {
			getData( block ) ^= mask;
		}

		bool containsUnexpected ( const unsigned long * const map ) const {
			bool result = ( map[0] & __data[0] ) == map[0];

			for ( uint_fast16_t block = 1;  block < __DD_BLOCK_COUNT__ && unlikely( result ); block++ ) {
				result = ( map[block] & getData( block ) ) == map[block];		
			} 

			return result;
		}
        
        static bool a_contains_b ( const __BIT_LIST_RAW_DATA_TYPE__ * const __a, const __BIT_LIST_RAW_DATA_TYPE__ * const __b ) {
			__BIT_LIST_RAW_DATA_TYPE__ result;
            uint_fast16_t i = 0;
            
            do {
                result =  ~__a[i] & __b[i];
            } while ( !result && ++i < __DD_BLOCK_COUNT__ );

			return !result;            
        }  
        
		inline bool contains ( const BitList & __b ) const {
			return a_contains_b( __data, __b.__data );
		}
        
		inline bool contains ( const __BIT_LIST_RAW_DATA_TYPE__ * const __b ) const {
			return a_contains_b( __data, __b );
		}
        
        inline bool operator<= ( const __BIT_LIST_RAW_DATA_TYPE__ * const __a ) const {
			return a_contains_b( __a, __data );     
        } 
        
        static bool reduced_a_contains_b ( const __BIT_LIST_RAW_DATA_TYPE__ * const __a,
                                           const __BIT_LIST_RAW_DATA_TYPE__ * const __b,
                                           const __BIT_LIST_RAW_DATA_TYPE__ * const __r ) {
                                               
			bool result;
            uint_fast16_t i = 0;
            
            do {
                result = ( __a[i] | ( __b[i] & ( ~__r[i] | ( __r[i] & ( __b[i] << 1 ))) )) == __a[i];
            } while ( ++i < __DD_BLOCK_COUNT__ && result );

			return result;            
        }
		
		inline bool reduced_a_contains_b ( const BitList & __b, const BitList & __r ) const {
            return reduced_a_contains_b( __data, __b.__data, __r.__data );
		}		
        
        inline bool reduced_a_contains_b ( const __BIT_LIST_RAW_DATA_TYPE__ * __b, const BitList & __r ) const {
			return reduced_a_contains_b( __data, __b, __r.__data );          
        }
        
        inline void reset () {
            __data = NULL;
        }

		const std::string toString () const;
        const std::string toString ( const unsigned int length ) const;
        
        static const std::string toString ( const __BIT_LIST_RAW_DATA_TYPE__ * data, const uint_fast16_t length ) {
            std::stringstream output;	

            output << "[";

            for ( uint_fast16_t pos = 0; pos < length; ++pos ) {
                if ( data[ pos / 64 ] & ( 1ul << ( pos % 64 ) )) {
                    output << "1";
                } else {
                    output << "0";
                }
            }
            output << "]";

            return output.str();            
        }
	};
}

#endif // __DD_BIT_LIST__
