/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_BIT_SELECTION_STRATEGY_FACTORY__
#define __DD_BIT_SELECTION_STRATEGY_FACTORY__

#include "DDOptions.h"

#include "DDRayWrapperContainer.h"
#include "DDBitCountVisitor.h"

#include "DDMostCommonBitFirstStrategy.h"
#include "DDMostAverageBitFirstStrategy.h"
#include "DDLeastAverageBitFirstStrategy.h"
#include "DDLeastCommonBitFirstStrategy.h"
#include "DDFirstAvailableBitStrategy.h"
#include "DDNarrowQueryBitsNeutralizationStrategy.h"
#include "DDNarrowInvQueryBitsNeutralizationStrategy.h"
#include "DDGreatestPosNegDiffFirstStrategy.h"

#include "DDVerQueryBitsNeutralizationStrategy.h"
#include "DDVerMostCommonBitFirstStrategy.h"
#include "DDVerLeastCommonBitFirstStrategy.h"

namespace addibit {

    template<class RAY_TYPE>
    class BitSelectionStrategyFactory {
        
        private:
        const uint_fast16_t mPlane;
        BitCountVisitor<RAY_TYPE> * mBitCounter;
        const uint_fast16_t mBitCounterSize;        
        
        uint_fast16_t get_counter_index ( const uintptr_t key ) const {
            uint_fast16_t c_index = 0;
            
            while ( c_index < mBitCounterSize && mBitCounter[ c_index ].key() != key ) {
                c_index++;
            }
            return c_index;
        }            
        
        inline uint_fast16_t plane () const {
            return mPlane;
        }
              
        const std::vector<uint_fast16_t> get_stats ( const RayWrapperContainer<RAY_TYPE> & container ) {
            const uintptr_t key = (uintptr_t) &container;
            uint_fast16_t c_index = get_counter_index( key );
            
            if ( c_index < mBitCounterSize ) {
                return mBitCounter[ c_index ].stats();
            } else {
                c_index = get_counter_index( 0 );
                container.iterate( mBitCounter[ c_index ] );
                mBitCounter[ c_index ].set_key( key );
                return mBitCounter[ c_index ].stats();
            }
        }            
        
        const BitSelectStrategy * createStrategy ( const RayWrapperContainer<RAY_TYPE> & container,
                                                   const BP_Tree_BitSelectionStrategy choice ) {
            switch ( choice ) {
                case BP_TREE_MOST_COMMON_BIT_FIRST:
                return new MostCommonBitFirstStrategy( get_stats( container ), container.size(), plane() );

                case BP_TREE_LEAST_COMMON_BIT_FIRST:
                return new LeastCommonBitFirstStrategy( get_stats( container ), container.size(), plane() );

                case BP_TREE_MOST_AVERAGE_BIT_FIRST:
                return new MostAverageBitFirstStrategy( get_stats( container ), container.size(), plane() );

                case BP_TREE_LEAST_AVERAGE_BIT_FIRST:
                return new LeastAverageBitFirstStrategy( get_stats( container ), container.size(), plane() );

                default:
                return new FirstAvailableBitStrategy( plane() );
             }
         }

        public:
        BitSelectionStrategyFactory ( const uint_fast16_t plane,
                                      const uint_fast16_t containers ) :
            mPlane( plane ),
            mBitCounter( new BitCountVisitor<RAY_TYPE>[ containers ] ),
            mBitCounterSize( containers ) {
                
            for ( uint_fast16_t index = 0; index < containers; ++index ) {
                mBitCounter[ index ].set_capacity( plane );
            }
        }

        ~BitSelectionStrategyFactory () {
            delete[] mBitCounter;
        }

        const BitSelectStrategy * createNarrowingStrategy ( const RayWrapperContainer<RAY_TYPE> & list_container,
                                                            const RayWrapperContainer<RAY_TYPE> & tree_container,
                                                            const BP_Tree_BitSelectionStrategy choice ) {
            
            switch ( choice ) {
                
                case BP_TREE_QUERY_BITS_NEUTRALIZATION:
                return new NarrowQueryBitsNeutralizationStrategy( get_stats( list_container ), list_container.size(), get_stats( tree_container ), tree_container.size(), plane() );    
            
                case BP_TREE_INV_QUERY_BITS_NEUTRALIZATION:
                return new NarrowInvQueryBitsNeutralizationStrategy( get_stats( list_container ), list_container.size(), get_stats( tree_container ), tree_container.size(), plane() );   
                
                case BP_TREE_GREATEST_POS_NEG_DIFF_FIRST:
                return new GreatestPosNegDiffFirstStrategy( get_stats( list_container ), list_container.size(), get_stats( tree_container ), tree_container.size(), plane() );
            
                default:
                return createStrategy( list_container, choice );
            }
        }
        
        const BitSelectStrategy * createVerificationStrategy ( const RayWrapperContainer<RAY_TYPE> & ver_container,
                                                               const RayWrapperContainer<RAY_TYPE> & list_container,
                                                               const RayWrapperContainer<RAY_TYPE> & tree_container,
                                                               const BP_Tree_BitSelectionStrategy choice ) {
            
            switch ( choice ) {
                
                case BP_TREE_QUERY_BITS_NEUTRALIZATION:
                return new VerQueryBitsNeutralizationStrategy( get_stats( ver_container ), ver_container.size(), get_stats( list_container ), list_container.size(), get_stats( tree_container ), tree_container.size(), plane() );
                
                case BP_TREE_MOST_COMMON_BIT_FIRST:
                return new VerMostCommonBitFirstStrategy( get_stats( list_container ), list_container.size(), get_stats( tree_container ), tree_container.size(), plane() );

                case BP_TREE_LEAST_COMMON_BIT_FIRST:
                return new VerLeastCommonBitFirstStrategy( get_stats( list_container ), list_container.size(), get_stats( tree_container ), tree_container.size(), plane() );
                
                default:
                return createStrategy( ver_container, choice );
            }
        }                
    };
}

#endif /* __DD_BIT_SELECTION_STRATEGY_FACTORY__ */
