/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_BIT_SELECTOR__
#define __DD_BIT_SELECTOR__

#include <stack>
#include <vector>

#include "DDBitList.h"
#include "DDBitSelectStrategy.h"
#include "DDInterval.h"

namespace addibit {

    class BitSelector {

        private:
        unsigned int mCurrent;

        protected:
        const std::vector<unsigned int> mMapping;
        std::stack<unsigned int> mHistory;

        virtual bool select( unsigned int bit, const BitList & map ) const {

            return bit < mMapping.size() - 1 && !map.isSet( mMapping[ bit ] );
        }

        public:
        BitSelector ( const BitSelectStrategy & strategy ) :
            mCurrent( 0 ),
            mMapping( strategy.createMapping() ) {}

        const BitList getFirstNBits ( const unsigned int n ) const {
            const Interval range( 0, n - 1 );

            return getBitsRange ( range );
        }

        const BitList getLastNBits ( const unsigned int n ) const {
            const Interval range( mMapping.size() - n - 1, mMapping.size() - 1 );

            return getBitsRange ( range );
        }

        const BitList getBitsRange ( const Interval & range ) const {
            BitList result( __BIT_LIST_ALLOC_SPACE__, 0 );

            for ( uint_fast16_t bit = range.lb(); bit <= range.ub(); bit++ ) {
                result.setBit( mMapping[ bit ] );
            }

            return result;
        } 
               
        unsigned int getNextBit () {
            mHistory.push( mCurrent );

            return mCurrent < mMapping.size() ? mMapping[ mCurrent++ ] : mMapping.size();            
        }

        BitList getNextNBits ( const uint_fast16_t n ) {
            BitList result( __BIT_LIST_ALLOC_SPACE__, 0 );
            
            for ( uint_fast16_t b_index = 0; b_index < n && checkNextBit(); ++b_index ) { 
                result.setBit( getNextBit() );
            }

            return result;
        }
 
        unsigned int getNextBit ( const BitList & map ) {
            mHistory.push( mCurrent );

            while ( select( mCurrent, map ) ) {
                mCurrent++;
            }

            return mCurrent < mMapping.size() ? mMapping[ mCurrent++ ] : 0;
        }      

        unsigned int getNextBit ( const BitList & map, const Interval & constraint ) {
            mHistory.push( mCurrent );

            while ( select( mCurrent, map & getBitsRange( constraint ) ) ) {
                mCurrent++;
            }

            return mCurrent < mMapping.size() ? mMapping[ mCurrent++ ] : 0;
        }   
        
        bool checkNextBit () const {
            return mCurrent < mMapping.size();
        }

        bool checkNextBit ( const BitList & map ) const {
            unsigned int bit = mCurrent;

            while ( select( bit, map ) ) {
                bit++;
            }            

            return bit < mMapping.size();
        }

        bool checkNextBit ( const BitList & map, const Interval & constraint ) const {
            unsigned int bit = mCurrent;

            while ( select( bit, map & getBitsRange( constraint ) ) ) {
                bit++;
            }            

            return bit < mMapping.size();
        }
 
        void revert () {
            mCurrent = mHistory.top();
            mHistory.pop();
        }

        unsigned int getTotalBits () const {
                return mMapping.size();
        }
    };
}

#endif // __DD_BIT_SELECTOR__
