/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_CONTAINER_VANTAGE_POINT_MAPPING__
#define __DD_CONTAINER_VANTAGE_POINT_MAPPING__

#include <vector>

#include "DDInterval.h"

namespace addibit {

    class ContainerVantagePointEntry {
        RayReference __vantage_point;
        uint_fast16_t * __density;       
 
        Interval __distance_range;
        Interval __inverse_range;
        
        public:
        ContainerVantagePointEntry () :
            __density( NULL ) {}
        
        ~ContainerVantagePointEntry () {
            if ( __density )
                delete[] __density;
        } 

        inline void set_vantage_point ( RayReference && vp ) {
            __vantage_point = std::move( vp );
            __density = new uint_fast16_t[ __vantage_point.degeneracy + 1 ];
            
            for ( uint_fast16_t i = 0; i <= __vantage_point.degeneracy; ++i ) {
                __density[i] = 0;
            } 
        }
            
        inline const RayReference & vantage_point () const {
            return __vantage_point;
        }
        
        inline const Interval & distance_range () const {
            return __distance_range;
        }
        
        inline const Interval & inverse_range () const {
            return __inverse_range;
        }
        
        uint_fast16_t update ( const RayReference & ref ) {
            uint_fast16_t dist = __vantage_point.activeSet.getCommonPopCnt( ref.activeSet );
            __distance_range.update( dist );
            __density[ dist ]++;
            return dist;
        }
        
        uint_fast16_t update ( RayReference && ref ) {
        
            uint_fast16_t dist = 0;
        
            if ( unlikely( __vantage_point.activeSet.data() == NULL ) ) {
                set_vantage_point( std::move( ref ) );
            } else {
                dist = __vantage_point.activeSet.getCommonPopCnt( ref.activeSet );
                __distance_range.update( dist );
                __inverse_range.update( ref.degeneracy - dist );
                __density[ dist ]++;
            }
            
            return dist;
        }
       
        uint_fast16_t get_distance ( const uint_fast32_t elements ) const {
            uint_fast16_t dist = 0;

            for ( uint_fast32_t local = __density[ 0 ]; dist < __vantage_point.degeneracy - 1 && local < elements; ++dist ) {
                local += __density[ dist + 1 ];
            } 
             
            return dist;
        }
 
        inline void init () {
            reset();
        }
        
        inline void reset () {
            __vantage_point = RayReference();
            if ( __density ) {
                delete[] __density;
                __density = NULL;
            }
            __distance_range.reset();
            __inverse_range.reset();
        }
    };

    class ContainerVantagePointMapping : public RayRefVisitor {
    
        private:
        ContainerVantagePointEntry * const __entries;
        ContainerVantagePointEntry * __end_ptr;
        
        uint_fast16_t * __ref_dist;
        
        std::vector<Interval> __conditional_dist;
        std::vector<Interval> __conditional_inv;
        
       static uint_fast32_t max_entries ( const uint_fast16_t max_depth, const uint_fast16_t split_factor ) {
            uint_fast32_t max_nodes = 1;
                
            for ( uint_fast16_t level = 1; level <= max_depth; ++level ) {
                max_nodes += pow( split_factor, level );
            }
            
            return max_nodes;
        }
        
        public:
        ContainerVantagePointMapping ( const uint_fast16_t max_depth, const uint_fast16_t split_factor ) :
            __entries( new ContainerVantagePointEntry[ max_entries( max_depth, split_factor ) ] ),
            __end_ptr( __entries ),
            __ref_dist( NULL ) {}
        
        ~ContainerVantagePointMapping () {
            delete[] __entries;
            if ( __ref_dist != NULL ) {
                delete[] __ref_dist;
            }
        }
        
        inline void init ( const uint_fast32_t ref_count ) {
            
            if ( __ref_dist != NULL ) {
                delete[] __ref_dist;
            }
            __ref_dist = new uint_fast16_t[ ref_count ];
        }
        
        inline uint_fast16_t last_ref_dist ( const uint_fast32_t id ) const {
            return __ref_dist[ id ];
        }
        
        inline void insert_empty () {
            ( __end_ptr++ )->init();
            
            __conditional_dist.emplace_back();
            __conditional_inv.emplace_back();
        }        
        
        inline const Interval & conditional_dist ( const uint_fast16_t id ) const {
            return __conditional_dist[ id ];
        }
        
        inline const Interval & conditional_inv ( const uint_fast16_t id ) const {
            return __conditional_inv[ id ];
        }       
        
        inline void update ( const uint_fast32_t id,
                             RayReference && ref ) { 
            
            __conditional_dist[ id ].update( __ref_dist[ ref.id ] );
            __conditional_inv[ id ].update( ref.degeneracy - __ref_dist[ ref.id ] );
            __ref_dist[ ref.id ] = __entries[ id ].update( std::move( ref ) );
        }            
        
        inline const ContainerVantagePointEntry & operator[] ( const uint_fast32_t id ) const {
            return __entries[ id ];
        }
        
        inline ContainerVantagePointEntry & operator[] ( const uint_fast32_t id ) {
            return __entries[ id ];
        }
        
        void reset () {
            while ( __end_ptr != __entries ) {
                ( --__end_ptr )->reset();
            }
        }
        
        void onVisit ( const RayReference & ref ) {
            __ref_dist[ ref.id ] = __entries[ 0 ].update( ref );
        }
        
        const uint_fast16_t * ref_dist () const {
            return __ref_dist;
        }
    };
}

#endif /* __DD_CONTAINER_VANTAGE_POINT_MAPPING__ */
