/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_DUAL_SET_RAY__
#define __DD_DUAL_SET_RAY__

#include "DDBitList.h"
#include "DDRay.h"

#include "DDLinearDependenceEntry.h"
#include "DDLinearDependenceCache.h"

namespace addibit {

	class InternalDependenceMap {

		private:
		std::vector<LinearDependenceEntry*> mCacheMap;
		unsigned char* mIndexToCoeff;

		public:
		InternalDependenceMap ( const unsigned char size ) :
			mCacheMap( 1, static_cast<LinearDependenceEntry*>( NULL )) {
			mIndexToCoeff = new unsigned char[size];

			for ( unsigned char i = 0; i < size; i++ ) {
				mIndexToCoeff[i] = 0;
			}
		}
		~InternalDependenceMap () {
			delete[] mIndexToCoeff;
		}

		const LinearDependenceEntry* getEntry ( const unsigned char index ) const {
			return mCacheMap[mIndexToCoeff[index]];
		}

		void setEntry ( LinearDependenceEntry* entry, const unsigned char index ) {
			mIndexToCoeff[index] = mCacheMap.size();
			mCacheMap.push_back( entry );
		}
	};

	class DualSetRay : public Ray {

		private:
		const BitList mPrimaryActiveSet;
		BitList mSecondaryActiveSet;
		
		LinearDependenceEntry ** const mInternalMap;

		public:

		DualSetRay ( mpz_class* const coeffs, LinearDependenceEntry** const internalDependenceMap, const BitList & primaryActiveSet, const BitList & secondaryActiveSet ) :
			Ray( coeffs ),
			mPrimaryActiveSet( primaryActiveSet ),
			mSecondaryActiveSet( primaryActiveSet | secondaryActiveSet ),
			mInternalMap( internalDependenceMap ) {}

		DualSetRay ( const BitList & primaryActiveSet, const BitList & secondaryActiveSet, LinearDependenceEntry** const internalDependenceMap, const uint_fast16_t plane, const uint_fast16_t total_planes ) :
			Ray( plane + 1, total_planes - ( plane + 1 ) ),
			mPrimaryActiveSet( primaryActiveSet ),
			mSecondaryActiveSet( secondaryActiveSet ),
			mInternalMap( internalDependenceMap ) {
        
            mSecondaryActiveSet |= primaryActiveSet;
        }

		~DualSetRay () {

			for ( uint_fast16_t block = 0, prefix = 0; block < __DD_BLOCK_COUNT__; ++block, prefix += __BLOCK_SIZE__ ) {
				__BIT_LIST_RAW_DATA_TYPE__ block_data = mSecondaryActiveSet[ block ] ^ mPrimaryActiveSet[ block ];

				while ( block_data ) {
					const uint_fast16_t suffix = __builtin_ctzl( block_data );
					unsigned long mask = __SINGLE_BIT__ << suffix;
					const uint_fast16_t index = prefix + suffix;

                    // @FIXME Use shared pointers here!
                    mInternalMap[ index ]->dereference();
					block_data ^= mask;
				}
			}

			delete[] ( mInternalMap + GLOBAL_RAY_DIM ) ;
		}

		const BitList & getPrimaryActiveSet () const {
			return mPrimaryActiveSet;
		}

		const BitList & getSecondaryActiveSet () const {
			return mSecondaryActiveSet;
		}

        const BitList & getMinimalActiveSet () const {
            return mPrimaryActiveSet;
        }

		const BitList & getActiveSet () const {
			return mSecondaryActiveSet;
		}

		inline const LinearDependenceEntry * getLinearDependenceEntry ( const uint_fast16_t index ) const {
			return mInternalMap[ index ];
		} 

		inline LinearDependenceEntry * link ( const uint_fast16_t index ) const {
            return mInternalMap[ index ]->reference();
		} 

		inline void setLinearDependenceEntry ( const uint_fast16_t index, LinearDependenceEntry * const entry ) {
			mInternalMap[ index ] = entry;
			mSecondaryActiveSet.setBit( index );
		} 

        inline bool degenerate () const {
            return mSecondaryActiveSet == mPrimaryActiveSet;
        }
        
        uint_fast16_t deg_degree () const {
            return mSecondaryActiveSet.getPopulationCount() - GLOBAL_RAY_MIN_FACETS;
        }

		const std::string toString () const;
	};
}

#endif // __DD_DUAL_SET_RAY__
