
#ifndef __DD_DYNAMIC_SAMPLE_CREATOR__
#define __DD_DYNAMIC_SAMPLE_CREATOR__

#include <vector>
#include <algorithm>

#include "DDBitList.h"
#include "DDSampleCreator.h"
#include "DDSamplesContainer.h"

namespace addibit {

    class DynamicSampleCreator : public SampleCreator {

        private:
        const float mIntensityFactor;
        const float mSampleDifference;
        const float mStretch;

        unsigned long getSum ( const std::vector<struct PopulationEntry> & stats ) const {
            unsigned long sum_total = 0;

            for ( unsigned int i = 0; i < stats.size(); i++ ) { 
                sum_total += stats[i].count;
            }            

            return sum_total;
        }

        const BitList getRegularBasicPattern ( const std::vector<struct PopulationEntry> & stats ) const {

            BitList result( mBlocks );
            unsigned long sum_total = getSum( stats );
            unsigned long sum = 0;

            std::vector<struct PopulationEntry> sorted_stats = stats; 
            sort( sorted_stats.begin(), sorted_stats.end() );

            for ( unsigned int i = 0; sum < sum_total * mIntensityFactor && i < sorted_stats.size(); i++, sum += sorted_stats[i].count ) { 
                result.setBit( sorted_stats[i].position );
            }

            return result;
        }

        void fillSamplesContainer ( SamplesContainer & container,
                                    const std::vector<struct RayReference>& entries,
                                    const unsigned int plane ) const {

             std::vector<struct PopulationEntry> stats = createStats( entries, plane );

             fillSamplesContainer( container, entries, stats, plane );
        }

        void fillSamplesContainer ( SamplesContainer & container,
                                    const std::vector<struct RayReference>& entries,
                                    std::vector<struct PopulationEntry> & stats,
                                    const unsigned int plane ) const {

            BitList current_reg( mBlocks );
            reorder( stats );

            /* Set up the initial state. */
            unsigned int current_index = 0;

            while ( current_index < mRegularThreshold ) {
                current_reg.setBit( stats[current_index++].position );
            }

            /* Create a sample so that at least 10% of the maps are in state 'over'. */     
            unsigned long current_limit = entries.size() * mIntensityFactor;

            do {
                container.insert( current_reg, getInverseBasicPattern( current_reg, plane ) );

                while ( getOverCount( container, entries ) < current_limit ) {
                    current_reg.setBit( stats[current_index++].position );
                    container.setLast( current_reg, getInverseBasicPattern( current_reg, plane ) );
                }
    
                current_limit += entries.size() * mSampleDifference + 1;
            } while ( current_limit < entries.size() * mStretch );
        }

        protected:
        
        virtual void reorder ( std::vector<struct PopulationEntry> & stats ) const = 0;

        public:
        DynamicSampleCreator( const unsigned int blocks,
                              const unsigned int minCommonBits,
                              const float intensityFactor,
                              const float sampleDifference,
                              const float stretch ) :
            SampleCreator( blocks, minCommonBits ),
            mIntensityFactor( intensityFactor ),
            mSampleDifference( sampleDifference ),
            mStretch( stretch ) {}

        const SamplesContainer createSamples( const std::vector<struct RayReference>& entries,
                                              const unsigned int plane,
                                              const unsigned int maxSamples ) const {

            SamplesContainer container( maxSamples, mRegularThreshold, mInverseThreshold );

            fillSamplesContainer( container, entries, plane );

            return container;
        } 

        const SamplesContainer createSamples( const std::vector<struct RayReference>& entries,
                                              std::vector<struct PopulationEntry> & stats,
                                              const unsigned int plane,
                                              const unsigned int maxSamples ) const {

            SamplesContainer container( maxSamples, mRegularThreshold, mInverseThreshold );
            fillSamplesContainer( container, entries, stats, plane );

            return container;
        } 

        const SamplesContainer createSamples( const std::vector<struct RayReference>& posEntries,
                                              const std::vector<struct RayReference>& negEntries,
                                              const unsigned int plane,
                                              const unsigned int maxSamples ) const {

            SamplesContainer container( maxSamples, mRegularThreshold, mInverseThreshold );
            fillSamplesContainer( container, posEntries, plane );
            fillSamplesContainer( container, negEntries, plane );

            return container;
        }        

        const SamplesContainer createSamples( const std::vector<struct RayReference>& posEntries,
                                              const std::vector<struct RayReference>& negEntries,
                                              const std::vector<struct RayReference>& zerEntries,
                                              const unsigned int plane,
                                              const unsigned int maxSamples ) const {

            SamplesContainer container( maxSamples, mRegularThreshold, mInverseThreshold );
            fillSamplesContainer( container, posEntries, plane );
            fillSamplesContainer( container, negEntries, plane );
            fillSamplesContainer( container, zerEntries, plane );

            return container;
        }  
    };
}

#endif // __DD_DYNAMIC_SAMPLE_CREATOR__
