/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_EDGE_CANDIDATES_CONTAINER__
#define __DD_EDGE_CANDIDATES_CONTAINER__

#include "DDAdjacencyCandidate.h"
#include "DDAdjacencyTestStack.h"
#include "DDEdgeCandidatesContainerVisitor.h"

namespace addibit {

    class EdgeCandidatesContainer {

        private:
        const uint_fast16_t __segment_size;
        const uint_fast32_t __memory_size;
        char * __memory; 
        
        AdjacencyTestStack * __stack;
        NarrowingCandidate * __current_candidate;
        
        char * const __start_ptr;
        char * const __end_ptr;
        char * __current_ptr;
        
        inline uint_fast32_t mem_size ( const uint_fast32_t capacity,
                                        const uint_fast16_t stack_size ) const {
                                            
            return __segment_size * capacity +
                   NarrowingCandidate::dynamic_size() +
                   stack_size + ( stack_size % 16 ); /* For aligning purposes. */
        }
        
        inline char * __alloc ( const uint_fast32_t size, const uint_fast16_t back_alignment = 8 ) {
            char * __address = __memory + ( size % back_alignment );
            __memory += size + ( size % back_alignment );
            return __address;
        } 
        
        public:
        EdgeCandidatesContainer ( const uint_fast32_t capacity,
                                  const uint_fast16_t stack_size ) :   
            __segment_size( AdjacencyCandidate::dynamic_size() ),
            __memory_size( mem_size( capacity, stack_size ) ),
            __memory( new ( aligned_alloc( 16, __memory_size )) char[ __memory_size ] ),
            __stack( new ( __alloc( stack_size, 16 ) ) AdjacencyTestStack() ),
            __current_candidate( new ( __alloc( NarrowingCandidate::dynamic_size() ) ) NarrowingCandidate() ),
            __start_ptr( __alloc( capacity * __segment_size ) ),
            __end_ptr( __memory ),
            __current_ptr( __start_ptr ) {}
            
        ~EdgeCandidatesContainer () {
            free( __memory - __memory_size );
        }
        
        inline void load ( const BitList & cand, const BitList & mask ) {
            __current_candidate->load( cand, mask );
        }
        
        inline const NarrowingCandidate * candidate () const {
            return __current_candidate;
        }
        
        inline NarrowingCandidate * candidate () {
            return __current_candidate;
        }
        
        inline const NarrowingCandidate & candidate_ref () const {
            return *__current_candidate;
        }
        
        inline const AdjacencyTestStack * stack () const {
            return __stack;
        }
        
        inline AdjacencyTestStack * stack () {
            return __stack;
        }        

        void clear () {
            __current_ptr = __start_ptr;
        }

        inline bool empty () const {
            return __current_ptr == __start_ptr;
        }

        inline uint_fast32_t size () const {
            return ( __current_ptr - __start_ptr ) / __segment_size;
        } 
        
        inline uint_fast32_t capacity () const {
            return ( __end_ptr - __start_ptr ) / __segment_size;
        }
        
        inline AdjacencyCandidate * push () {
            AdjacencyCandidate * __address = reinterpret_cast<AdjacencyCandidate *>( __current_ptr );
            __current_ptr += __segment_size;
            return __address;
        }
        
        void push ( const __BIT_LIST_RAW_DATA_TYPE__ * const entry, const uintptr_t ray_address ) {
            AdjacencyCandidate * __address = reinterpret_cast<AdjacencyCandidate *>( __current_ptr );
            __address->set_id( ray_address );
            __address->cut_and_save( __current_candidate->data(), entry );
            __current_ptr += __segment_size;
        }
        
        inline const AdjacencyCandidate * pop () {
            __current_ptr -= __segment_size;
            return reinterpret_cast<AdjacencyCandidate *>( __current_ptr );
        }            
        
        template<class RAY_TYPE, class VISITOR>
        void verify ( const RAY_TYPE * const p_1, VISITOR & visitor ) {
            while ( !empty() ) {
                visitor.verify_edge( p_1, pop() );
            }
        }        
    };
}

#endif // __DD_EDGE_CANDIDATES_CONTAINER__
