/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_EXTENDED_BIT_PATTERN_CONDITION__
#define __DD_EXTENDED_BIT_PATTERN_CONDITION__

#include "DDTreeNodeCondition.h"

namespace addibit {

    union NarrowingResult {
        __m64 __mmx;
        long __data;
    };

	class ExtendedBitPatternCondition : public TreeNodeCondition {
	
	    private:
	    uint_fast16_t mPopCnt;
		BitList mCutMap;
		
		public:
		ExtendedBitPatternCondition () {}
		
        ExtendedBitPatternCondition ( BitList && union_map,
                                      const uint_fast32_t pop_cnt,
                                      BitList && cut_map ) :
            TreeNodeCondition( std::move( union_map ) ),
            mPopCnt( pop_cnt ),
            mCutMap( std::move( cut_map ) ) {}

        ExtendedBitPatternCondition ( const ExtendedBitPatternCondition & lvalue ) :
            TreeNodeCondition( lvalue ),
            mPopCnt( lvalue.mPopCnt ),
            mCutMap( lvalue.mCutMap ) {}
        
        ExtendedBitPatternCondition ( ExtendedBitPatternCondition && rvalue ) :
            TreeNodeCondition( std::move( rvalue ) ),
            mPopCnt( rvalue.mPopCnt ),
            mCutMap( std::move( rvalue.mCutMap ) ) {}
        
        inline void operator= ( ExtendedBitPatternCondition && rvalue ) {
            TreeNodeCondition::operator= ( std::move( rvalue ) );
            mPopCnt = rvalue.mPopCnt;
            mCutMap = std::move( rvalue.mCutMap );
        }

        inline void operator= ( const ExtendedBitPatternCondition & lvalue ) {
            TreeNodeCondition::operator= ( lvalue );
            mPopCnt = lvalue.mPopCnt;
            mCutMap = lvalue.mCutMap;
        }
		
        inline bool proceed_narrowing ( const NarrowingCandidate & candidate ) const {

            return TreeNodeCondition::proceed_narrowing( candidate ) &&
                   mCutMap.getDiffPopCnt( candidate.data() ) <= mPopCnt - GLOBAL_EDGE_DIM;
        }		
        
        inline bool proceed_verification ( const EdgeCandidate & candidate ) const {
        
            return TreeNodeCondition::proceed_verification( candidate ) &&
                   mCutMap.getDiffPopCnt( candidate.data() ) <= mPopCnt - ( candidate.population() + 1 ); 
        }
        
        /**
         * node_data = { mPopCnt, union_map, cut_map }
         */
        void write_data ( __BIT_LIST_RAW_DATA_TYPE__ * data_region ) const {
            
            ( data_region++ )[0] = mPopCnt;
            uint_fast16_t i = 0;

            do {
                data_region[i] = mMap[i];
                data_region[i + __DD_BLOCK_COUNT__] = mCutMap[i];
            } while( ++i < __DD_BLOCK_COUNT__ );
        }
        
        inline void write_narrowing_data ( __BIT_LIST_RAW_DATA_TYPE__ * data_region ) const {
            write_data( data_region );
        }

        inline void write_verification_data ( __BIT_LIST_RAW_DATA_TYPE__ * data_region ) const {
            write_data( data_region );
        }
         
         uint_fast16_t mem_size () const {
             return ( 2 * __DD_BLOCK_COUNT__ + 1 ) * sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
         }
         
        inline static bool narrow_feasible ( EdgeCandidatesContainer & container, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::NARROWING_OPS++;
    #endif
            NarrowingResult res;
            
            res.__mmx = _mm_cmpgt_pi32( BitList::nar_pop_cnt_a_minus_b( node_data + 1, container.candidate()->data() ), _mm_set_pi32( GLOBAL_EDGE_DIM - 1, node_data[ 0 ] - GLOBAL_EDGE_DIM ) );
            
            return res.__data == 0xFFFFFFFF00000000;
        }
        
        inline static void intermediate_post_action ( EdgeCandidatesContainer & , const __BIT_LIST_RAW_DATA_TYPE__ * const ) {}
        inline static void leaf_post_action ( EdgeCandidatesContainer & , const __BIT_LIST_RAW_DATA_TYPE__ * const ) {}
        inline static bool intermediate_post_action ( EdgeCandidate &, const __BIT_LIST_RAW_DATA_TYPE__ * const ) {
            return false;
        }
        inline static bool leaf_post_action ( EdgeCandidate &, const __BIT_LIST_RAW_DATA_TYPE__ * const ) {
            return false;
        }
        
        inline static bool verify_feasible ( const EdgeCandidate & candidate, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
    
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::VERIFICATION_OPS++;
    #endif 
    
             return BitList::ver_pop_cnt_a_minus_b( node_data + 1, candidate.data(), node_data[ 0 ] - candidate.population() + 1 );
        }
        
        static uint_fast16_t dynamic_size () {
            return TreeNodeCondition::dynamic_size() + ( __DD_BLOCK_COUNT__ + 1 ) * sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
        
        static uint_fast16_t required_alignment () {
            return sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
        
        static uint_fast16_t guaranteed_alignment () {
            return sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
	};
}
	
	
#endif /* __DD_EXTENDED_BIT_PATTERN_CONDITION__ */
