/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_EXTREME_RAY_GENERATOR__
#define __DD_EXTREME_RAY_GENERATOR__

#include <list>
#include <boost/thread.hpp>

#include "DDMatrix.h"
#include "DDStats.h"
#include "DDLogging.h"

#include "DDBasisCacheEntryNode.h"
#include "DDTerminalCacheEntryNode.h"

#include "DDCacheFactory.h"

namespace addibit {

    template <class RAY_TYPE, class CACHE_ENTRY_TYPE>
	class ExtremeRayGenerator {
	
	    private:
        const uint_fast16_t mId;
        std::deque<CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * > mBuffer;

		protected:
		const Matrix & mSystem;
        std::list<boost::thread *> mSubthreads;        

        virtual const struct StepStats enumerate ( const unsigned long plane, CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * ) = 0;

		void calculate_step ( CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * const node ) {

            DD_START_OUTPUT ( STEP_LOG )

	        CACHE_ENTRY_TYPE & cache_entry = node->data();
            
	        DD_CONCAT_OUTPUT ( STEP_LOG,
                              "[" << node->id() << "] +/-/deg/narrowed/verified/extremal = " )

	        START_TIMER( STEP_TIME )

            const struct StepStats stats = enumerate( cache_entry.plane(), node ); 

            DD_CONCAT_OUTPUT ( STEP_LOG,
                               cache_entry.getPositiveCount() << "/" << cache_entry.getNegativeCount() << "/" << std::fixed << stats.degeneracy << "/" <<
                               stats.narrowed << "/" << stats.verified << "/" << stats.extremal )

	        STOP_TIMER( STEP_TIME )

	        DD_CONCAT_OUTPUT ( STEP_LOG, ", dt=" << TIMER_VALUE( STEP_TIME ) << "ms" )
            
        #ifdef __STATS_OUTPUT__
            DD_CONCAT_OUTPUT ( STEP_LOG, ", overhead=" << stats.tree_creation_time << "ms" )
        
            if ( Options::current().plot_required() ) {
                LOG_LINE( COMPLEXITY_PLOT, node->id() << " " << stats.narrowing_complexity << " " << stats.verification_complexity << " " << stats.degeneracy << " " << (double_t)TIMER_VALUE( STEP_TIME ) / (double_t)stats.narrowing_queries << " " << (double_t)TIMER_VALUE( STEP_TIME ) / (double_t)stats.nontrivial_candidates )
            }
        #endif            
            
            DD_FLUSH_OUTPUT ( STEP_LOG )
        }
        
        template<class GENERATOR>
        inline void t_fork ( CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * const node ) {
            GENERATOR generator( mSystem, node );
            boost::thread * gen_thread = new boost::thread( generator );
            mSubthreads.push_back( gen_thread ); 
        }        
		
		virtual void calculateBasis ( CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * const ) = 0;       
        virtual void fork ( CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * const ) = 0;
        
        public:
		ExtremeRayGenerator ( const Matrix & system ) :
            mId( 0 ),
	        mSystem( system ) {}            
	    
		ExtremeRayGenerator ( const Matrix & system,
		                      CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * const node ) :
            mId( node->thread_id() ),
	        mSystem( system ) {
            
            mBuffer.push_back( node );
        }	    
        
        ExtremeRayGenerator ( const ExtremeRayGenerator<RAY_TYPE, CACHE_ENTRY_TYPE> & other ) :
            mId( other.mId ),
            mBuffer( other.mBuffer ),
            mSystem( other.mSystem ) {
        }
        
        virtual ~ExtremeRayGenerator () {}
        
        void init ( CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * const head ) {
           
            /* Find a nonsingular basis matrix and calculate an initial cone to start the computation with. */
            calculateBasis( head );
            head->enum_successors( mBuffer );
            delete head;            
        }

        void operator() () {
            
            CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * head;
        
	        while ( !mBuffer.empty() ) {
	            head = mBuffer.front();
                mBuffer.pop_front();
                
                if ( likely( head->thread_id() == mId ) ) {

                    /* Intersect the current cone with the next hyperplane. */
                    head->prepare_for_test();
                    calculate_step( head );
                    head->enum_successors( mBuffer );
                    delete head;
                } else {
                    /* Start a new generator for this branch. */
                    fork( head );
                }
	        }
            
            for ( auto subthread : mSubthreads ) {
                subthread->join();
                delete subthread;
            }
        }
        
        const Matrix & system () const {
            return mSystem;
        }
    };
}

#endif // __DD_EXTREME_RAY_GENERATOR__
