/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_EXTREME_RAY_UNCOMPRESSOR__
#define __DD_EXTREME_RAY_UNCOMPRESSOR__

#include <gsl/gsl_linalg.h>

#include "DDCalculationMatrix.h"
#include "DDSLESolver.h"
#include "DDMatrix.h"
#include "DDSingleSetRay.h"
#include "DDOutputVector.h"

#include "DDLogging.h"

namespace addibit {

	class ExtremeRayUncompressor : public SLESolver {

		public:
		ExtremeRayUncompressor ( const uint_fast16_t rows, const uint_fast16_t columns ) :
            SLESolver( rows, columns ) {}
            
		~ExtremeRayUncompressor () {}

		void load ( const Matrix & matrix, const BitList& rowMap ) {

			unsigned int dst_row = 0;

			for ( uint_fast16_t block = 0, prefix = 0; block < GLOBAL_BLOCKS; block++, prefix += __BLOCK_SIZE__ ) {
				unsigned long block_data = rowMap[block];

				while ( block_data ) {
					unsigned int suffix = __builtin_ctzl( block_data ); 
					unsigned int src_row = prefix + suffix;

					load( matrix, src_row, dst_row++ );

					block_data ^= ( __SINGLE_BIT__ << suffix );
				}
			}

			setRows( dst_row );
			setColumns( matrix.getColumns() );
		}

        void load ( const Matrix& matrix, const BitList& rowMap, double * dst ) const {
            
			unsigned int dst_row = 0;

			for ( uint_fast16_t block = 0, prefix = 0; block < GLOBAL_BLOCKS; block++, prefix += __BLOCK_SIZE__ ) {
				unsigned long block_data = rowMap[block];

				while ( block_data ) {
					unsigned int suffix = __builtin_ctzl( block_data ); 
					unsigned int src_row = prefix + suffix;

					load( matrix, src_row, dst_row++, dst );

					block_data ^= ( __SINGLE_BIT__ << suffix );
				}
			}
        }

		void load ( const Matrix& matrix, const unsigned int srcRow, const unsigned int dstRow ) {

			for ( unsigned int col = 0; col < matrix.getColumns(); col++ ) {
				setValue( dstRow, col, matrix[srcRow][col] );
			}
		}

		void load ( const Matrix& matrix, const unsigned int srcRow, const unsigned int dstRow, double * dst ) const {

			for ( unsigned int col = 0; col < matrix.getColumns(); col++ ) {
                dst[ dstRow*matrix.getColumns() + col ] = matrix[srcRow][col].get_d();
			}
		}

        template<class RAY>
		void calcExtremeRay ( const Matrix &, const RAY* ray, BitList & result ) {
			result = ray->getActiveSet();
		}

        template<class RAY>
		void calcExtremeRay ( const Matrix& matrix, const RAY* ray, OutputVector<double> & result ) {
			const BitList & row_map = ray->getActiveSet();

			calcExtremeRay( matrix, row_map, result );
		}

        void calcExtremeRay ( const Matrix &, const BitList &, OutputVector<double> & );

        template<class RAY>
		void calcExtremeRay ( const Matrix& matrix, const RAY* ray, RepresentationVector& result ) {
			const BitList & row_map = ray->getMinimalActiveSet();

			calcExtremeRay( matrix, row_map, result );
		}

		void calcExtremeRay ( const Matrix & matrix, const BitList & map, RepresentationVector & result ) {

            load( matrix, map );
			transformToUpperTriangularMatrix();
            cut_zero_rows();
#ifdef __MINIMIZE_MUL_OPERATIONS__
            simplifyUpperTriangularMatrix( getRows() );
#endif
            transformToDiagonalMatrix();
            prepare_for_output( getColumns() - 1 );
			calcExtremeRay( result );

		    if ( result * matrix[ map.getFirstUnset() ]  < 0 ) {
			    result.negate();
            }
		}

		void calcExtremeRay ( RepresentationVector& result ) const {

            if ( unlikely( empty() ) ) {
                result.setValue( 0, 1 );
            } else {
                result.setValue( getColumns() - 1, -getValue( 0, 0 ) );

                UtilsFactory::CALC_UTILS.startIncrementalGCD( result.getValue( getColumns() - 1 ) );

			    for ( unsigned int index = 0; index < getRows(); index++ ) {
                    const CalculationVector & current_row = getRow( index );

				    if ( current_row.isNonZeroAt( getColumns() - 1 ) && result.isNonZeroAt( getColumns() - 1 ) ) {
					    result.setValue( index, current_row.getValue( getColumns() - 1 ) );
                        UtilsFactory::CALC_UTILS.calcIncrementalGCD( result.getValue( index ) );
				    }
			    }

                result /= UtilsFactory::CALC_UTILS.getResult();
            }
		}
	};
}

#endif //  __DD_EXTREME_RAY_UNCOMPRESSOR__
