/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <cstdlib>

#include "DDInput.h"
#include "DDUtilsFactory.h"

#include "DDLexSortCriterion.h"
#include "DDCoLexSortCriterion.h"
#include "DDRevLexSortCriterion.h"
#include "DDRevCoLexSortCriterion.h"
#include "DDMinZeroFirstSortCriterion.h"
#include "DDMaxZeroFirstSortCriterion.h"

#include "DDLogging.h"
#include "DDOptions.h"

using namespace addibit;

Input::Input ( const char * file ) :
    __ordering( NO_ORDERING ),
	mSystem ( getMatrixFromFile ( file ) ),
	transformator ( Input::getIdentityMatrix( mSystem.getColumns() )),
	transformed ( mSystem ) {}

Input::Input ( const char * file, const InputOrdering ordering ) :
    __ordering( ordering ),
	mSystem ( getMatrixFromFile ( file ) ),
	transformator ( Input::getIdentityMatrix( mSystem.getColumns() )),
	transformed ( mSystem ) {}

const Matrix Input::getMatrixFromFile ( const char * file ) {
	std::ifstream input( file );
    std::string line;

//	input.open( file );
    input.exceptions(std::ifstream::eofbit);
	
	if ( unlikely( !input ) ) {
		throw "Can not open file!";
	}

    try {
    /* Search for the start of the matrix definition - 'begin'. */
        do {            
            std::getline( input, line );
            Options::remove_comments( line, '*' );
            Options::trim( line );        
        } while( line != "begin" );

        /* Read the matrix. */
        Matrix system = getMatrixFromFile( input );

        /* Check whether the matrix was correctly bounded. */
        input >> line;

        if ( unlikely( line != "end" ) ) {
            DD_OUTPUT( "Matrix file seems to be incorrect. 'end' was expected but something different was found: " << line )
            return Matrix( 0, 0 );
        }

        process_options( system, input );
        reorder_for_computation( system );        
         
        reorder_for_basis( system );
        
        /* Small trick for rectangular matrices as the algorithm assumes rows > columns. */
        if ( system.is_rectangular() ) {
            system.duplicate_row( 0 );
        }

        return system;
    } catch ( std::ifstream::failure e ) {
        DD_OUTPUT( "Something happened while parsing the input file!" )
        return Matrix( 0, 0 );
    }
}

void Input::process_options ( Matrix & system, std::ifstream & input ) {

    std::string line;
    std::vector<uint_fast16_t> strict_rows;
    std::vector<uint_fast16_t> eq_rows;
    input.exceptions( std::ifstream::badbit );
     
    while ( input >> line ) {
        
        if ( line == "equality" ) {
            uint_fast16_t number;
            input >> number;   

            for ( uint_fast16_t i = 0; i < number; ++i ) {
                int index;
                input >> index;
                eq_rows.push_back( system.getColumns() + index );
            } 
        } else if ( line == "strict_inequality" ) {
            uint_fast16_t number;
            input >> number;   

            for ( uint_fast16_t i = 0; i < number; ++i ) {
                int index;
                input >> index;
                strict_rows.push_back( system.getColumns() + index );
            } 
        } else if ( line == "lexmin" ) {
            __ordering = LEX;
        } else if ( line == "lexmax" ) {
            __ordering = REV_LEX;
        } else if ( line == "colexmin" ) {
            __ordering = COLEX;
        } else if ( line == "colexmax" ) {
            __ordering = REV_COLEX;
        } else if ( line == "zeromin" ) {
            __ordering = MIN_ZERO;
        } else if ( line == "zeromax" ) {
            __ordering = MAX_ZERO;
        }
    }
    
    system.mark( strict_rows, eq_rows );   
}

const Matrix Input::getMatrixFromFile ( std::ifstream & input ) const {
	unsigned long dimension, inequalities;
	std::string value;
    std::string type;

	input >> inequalities;
	input >> dimension;
    input >> type;

	std::vector<unsigned long> order( inequalities );

	for ( unsigned long row = 0; row < inequalities; row++ ) {
        order[ row ] = row;
	} 

	Matrix system ( inequalities, dimension );

	for ( unsigned long row = 0; row < inequalities; row++ ) {

        if ( type == "rational" ) {
            std::vector<mpq_class> row_vec( dimension );

	        for ( unsigned long column = 0; column < dimension; column++ ) {
                input >> value;
	            row_vec[ column ] = mpq_class( value );

                if ( unlikely( column == 0 ) ) {                   
                    UtilsFactory::CALC_UTILS.startIncrementalLCM( row_vec[ column ].get_den() );
                } else {
                    UtilsFactory::CALC_UTILS.calcIncrementalLCM( row_vec[ column ].get_den() );
                }
		    }

	        UtilsFactory::CALC_UTILS.convertResultToAbsoluteValue();

	        for ( unsigned long column = 0; column < dimension; column++ ) {
                system.setValue( order[ row ], column, ( UtilsFactory::CALC_UTILS.getResult() / row_vec[ column ].get_den() ) * row_vec[ column ].get_num() );
		    }

        } else if ( type == "integer" ) {

	        for ( unsigned long column = 0; column < dimension; column++ ) {
	            input >> value;
			    system.setValue( order[ row ], column, value );
		    }
        }
	}

	return system;
}

void Input::reorder_for_basis ( Matrix & system ) const {
    
    uint_fast16_t row;
    
	RankCalculator calculator( system.getRows(), system.getColumns() );
    calculator.load_first_row( system );
    
    for ( row = 1; row < system.equalities(); ++row ) {
        calculator.addRow( system.row( row ) );
    }

	for ( ; row < system.getColumns(); ++row ) {

        calculator.addRow( system.row( row ) );
		bool linear_independent = calculator.hasMinimumRank( row + 1 );
        
		if ( !linear_independent ) {
			for ( uint_fast16_t sec_row = system.getColumns(); sec_row < system.getRows() - system.strict() && !linear_independent; sec_row++ ) {
				system.swapRows( row, sec_row );
                calculator.overwriteLastRow( system.row( row ) );
                linear_independent = calculator.hasMinimumRank( row + 1 );
			}
		}
        
		if ( unlikely( !linear_independent ) ) {
			DD_OUTPUT( "Cannot extract a base with rank " << ( row + 1 ) )
			exit( -1 );
		} 
	}
}

void Input::reorder_for_computation ( Matrix & system ) const {

    switch ( __ordering ) {
        case LEX:
            system.sort( LexSortCriterion() );
            break;
        case COLEX:
            system.sort( CoLexSortCriterion() );
            break;            
        case REV_LEX:
            system.sort( RevLexSortCriterion() );
            break;
        case REV_COLEX:
            system.sort( RevCoLexSortCriterion() );
            break;        
        case MIN_ZERO:
            system.sort( MinZeroFirstSortCriterion() );
            break;
        case MAX_ZERO:
            system.sort( MaxZeroFirstSortCriterion() );
            break;
     /*   default:
            system.sort( LexSortCriterion() ); */
    }
}

const Matrix Input::applyPositiveBaseTransformation ( const Matrix & matrix ) const {
	Matrix base = matrix.getSubMatrix( 0, matrix.getColumns() );
	
	return applyNemhauserWolsey( base );
}

const Matrix Input::applyNemhauserWolsey ( Matrix& matrix ) const {

	Matrix C = Input::getIdentityMatrix( matrix.getRows() );
	Matrix transformation = Input::getIdentityMatrix( matrix.getRows() );
	unsigned long row, column;

	for ( row = 0; row < std::min( matrix.getRows(), matrix.getColumns() ); row++ ) {
	
		for ( column = row + 1; column < matrix.getColumns(); column++ ) {

			if ( matrix.isNonZeroAt( row, column ) ) {
				UtilsFactory::CALC_UTILS.calcExendedGCD ( matrix[row][row], matrix[row][column] );

				C.setValue ( row, row, UtilsFactory::CALC_UTILS.getRegisterS() );
				C.setValue ( column, row, UtilsFactory::CALC_UTILS.getRegisterT() );
				C.setValue ( row, column, -matrix[row][column]/UtilsFactory::CALC_UTILS.getResult() );
				C.setValue ( column, column, matrix[row][row]/UtilsFactory::CALC_UTILS.getResult() );

				matrix *= C;
				transformation *= C;

				/* Transform C back to identity matrix */
				C.setValue ( row, row, 1 );
				C.setValue ( column, row, 0 );
				C.setValue ( row, column, 0 );
				C.setValue ( column, column, 1 );
			}
		}
		
		if ( matrix.getValue( row, row ) < 0 ) {
			
			C.setValue ( row, row, -1 );
			
			matrix *= C;
			transformation *= C;	

			/* Transform C back to identity matrix */
			C.setValue ( row, row, 1 );
		}

		for ( column = 0; column < row; column++ ) {

			if ( matrix.isNonZeroAt( row, column ) ) {

				C.setValue ( row, column, -( matrix[row][column]/matrix[row][row] + 1 ) );

				matrix *= C;
				transformation *= C;

				/* Transform C back to identity matrix */
				C.setValue ( row, column, 0 );
			}
		} 

		for ( column = 0; column < row; column++ ) {

			if ( matrix.isNonZeroAt( row, column ) ) {

				if ( matrix[row][column] % matrix[row][row] != 0 ) {
					C.setValue ( row, column, ( abs( matrix[row][column]/matrix[row][row]) + 1 ) );
				} else {
					C.setValue ( row, column, abs( matrix[row][column]/matrix[row][row] ) );
				}

				matrix *= C;
				transformation *= C;

				/* Transform C back to identity matrix */
				C.setValue ( row, column, 0 );
			}
		} 
	}

	for ( row = 0; row < std::min( matrix.getRows(), matrix.getColumns() ); row++ ) {

		matrix.setValue( row, row, 1 );

		for ( column = 0; column < row; column++ ) {			

			if ( matrix.isNonZeroAt( row, column ) ) {

				C.setValue( row, column, -matrix[row][column] );

				matrix *= C;
				transformation *= C;

				/* Transform C back to identity matrix */
				C.setValue ( row, column, 0 );
			}
		}
	}	

	return transformation;	
}
