/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_LINEAR_DEPENDENCE_CACHE__
#define __DD_LINEAR_DEPENDENCE_CACHE__

#include <stack>
#include <vector>

#include "DDBitList.h"
#include "DDLinearDependenceVector.h"
#include "DDMatrix.h"
#include "DDSLESolver.h"

#include "DDLogging.h"

namespace addibit {

	class LinearDependenceCache {

		private:
		const Matrix & mSystem;
		SLESolver mSolver;

		bool mLoaded;

		std::vector<LinearDependenceVector> data;
		std::stack<LinearDependenceEntry*> entries;
		std::stack<unsigned long> mSecondaryBits;

		public:
		LinearDependenceCache ( const Matrix & system );

		void insert ( const unsigned long index, const BitList& map ) {
			data[index].insert( map );
		}

		void insert ( const unsigned long index, LinearDependenceEntry * const entry ) {
			data[index].insert( entry );
		}

		const LinearDependenceEntry * getDependenceEntry ( const unsigned long index, const BitList& map ) const {
			return data[index].getDependenceEntry( map );
		}

		const LinearDependenceEntry * getDependenceEntry ( const unsigned long plane, const unsigned long position ) const {
			return data[plane].getDependenceEntry( position );
		} 

		LinearDependenceEntry * getEmptyLinearDependenceEntry ( const BitList & primarySet, const unsigned long secondaryBit ) {

			LinearDependenceEntry * entry = new LinearDependenceEntry( secondaryBit + 1 ); 

			entries.push( entry );

			if ( !mLoaded ) {
				mSolver.load( mSystem, primarySet );
				mLoaded = true;
			}
			mSolver.load( mSystem, secondaryBit );

			return entry;
		}

		LinearDependenceEntry * acquireLinearDependenceEntry ( const BitList & primarySet, const unsigned long secondaryBit ) {

			const unsigned long index = secondaryBit - mSystem.getColumns();
			LinearDependenceEntry * dependence_entry = data[index].acquireDependenceEntry( primarySet );

			if ( !dependence_entry ) {
				dependence_entry = getEmptyLinearDependenceEntry( primarySet, secondaryBit );
				insert( index, dependence_entry );
				mSecondaryBits.push( index );
			}

            dependence_entry->reference();
				
			return dependence_entry;
		}

		const LinearDependenceEntry * getLinearDependenceEntry ( const BitList & primarySet, const unsigned long secondaryBit ) {

			const unsigned long index = secondaryBit - mSystem.getColumns();
			return getDependenceEntry( index, primarySet );
		}

		void update () {

			if ( mLoaded ) {
				mSolver.solve();

				mLoaded = false;

				do {
					mSolver.fillLinearDependenceEntry( entries.top() );
					insert( mSecondaryBits.top(), entries.top()->getMap() );
					mSecondaryBits.pop();
					entries.pop();
				} while ( !entries.empty() );
			}
		}

        void releaseUnusedReferences ( const unsigned int plane ) {

            for ( unsigned int i = 0; i < /*mSystem.getRows() */ plane - mSystem.getColumns(); i++ ) {
                data[ i ].releaseUnusedReferences();
            }
        }

		unsigned long getSize() const;

		const std::string toString () const;
	};
}

#endif // __DD_LINEAR_DEPENDENCE_CACHE__
