/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_MATRIX__
#define __DD_MATRIX__

#include <vector>
#include <cstring>
#include <string>
#include <gmp.h>
#include <gmpxx.h>

#include "DDTweaks.h"

#include "DDBitList.h"
#include "DDVector.h"
#include "DDSortCriterion.h"
#include "DDRepresentationVector.h"
#include "DDCalculationVector.h"
#include "DDUtilsFactory.h"

namespace addibit {

	class Matrix {

		private:
		std::vector<CalculationVector> mData;
		std::vector<uint_fast16_t> mMapping;
        uint_fast16_t __equalities;
        uint_fast16_t __strict;

        const std::string toPortaFormat( const mpz_class & val, bool first ) const;

		protected:
		CalculationVector & getRow ( const uint_fast16_t index ) {
			return mData[ mMapping[ index ] ];
		}

		public:
		Matrix ();
		Matrix ( const uint_fast16_t rows, const uint_fast16_t columns );
		Matrix ( const uint_fast16_t rows, const uint_fast16_t columns, const std::vector<uint_fast16_t> & mapping );
		Matrix ( const Matrix & other );
		
		virtual ~Matrix () {}

		uint_fast16_t getRows () const {
			return mData.size();
		}

		virtual uint_fast16_t getColumns () const {

            if ( unlikely( mData.empty() )) {
                return 0;
            } else {
			    return mData[0].getDimension();
            }
		}
        
        bool is_rectangular () const {
            return getRows() == getColumns();
        }

		void insert ( const CalculationVector & row );

		void insert ( const CalculationVector & row, const uint_fast16_t pos ) {
			mMapping.insert( mMapping.begin() + pos, mData.size() );
			mData.push_back( row );
		}
        
        void mark ( const std::vector<uint_fast16_t> & , const std::vector<uint_fast16_t> & );
        
        void duplicate_row ( const uint16_t index ) {
            insert( getRow( index ) );
        }
		
		void moveRow ( const uint_fast16_t current_index,
		               const uint_fast16_t new_index ) {
		      
		    mMapping.insert( mMapping.begin() + new_index, mMapping[ current_index ] );
		    if ( current_index > new_index ) {
		        mMapping.erase( mMapping.begin() + current_index + 1 );
		    } else {
		        mMapping.erase( mMapping.begin() + current_index );
		    }   
		}
		
		void moveRowToLast ( const uint_fast16_t index ) {		    
		    const uint_fast16_t map = mMapping[ index ];
		    mMapping.erase( mMapping.begin() + index );
		    mMapping.push_back( map );  
		}		
		
		void moveToBack ( const std::vector<uint_fast16_t> & rows ) {
		    uint_fast16_t offset = 0;
		    
		    for ( uint_fast16_t index : rows ) {
		        moveRowToLast( index - offset++ );
		    }
		}
        
        uint_fast16_t equalities () const {
            return __equalities;
        }
        
        uint_fast16_t strict () const {
            return __strict;
        }

		const CalculationVector & getRow ( const uint_fast16_t index ) const {
			return mData[mMapping[index]];
		}

		const CalculationVector & row ( const uint_fast16_t index ) const {
			return mData[mMapping[index]];
		}

		const CalculationVector& getLastRow () const {
			return getRow( getLastRowIndex() );
		}

		uint_fast16_t getLastRowIndex () const {
			return getRows() - 1;
		}

		bool isNonZeroAt ( const uint_fast16_t row, const uint_fast16_t column ) const {
			return getRow( row ).isNonZeroAt( column );
		}

		void setValue ( const uint_fast16_t row, const uint_fast16_t column, const mpz_class & value ) {
			getRow( row ).setValue( column, value );
		}

		void setValue ( const uint_fast16_t row, const uint_fast16_t column, const long value ) {
			getRow( row ).setValue( column, mpz_class( value ));
		}

		void setValue ( const uint_fast16_t row, const uint_fast16_t column, const std::string & value ) {
			getRow( row ).setValue( column, mpz_class( value ));
		}

		const mpz_class& getValue ( const uint_fast16_t row, const uint_fast16_t column ) const {
			return getRow( row ).getValue( column );
		}

		const CalculationVector& operator[] ( const uint_fast16_t index ) const {
			return getRow( index );
		}

		void operator = ( const Matrix& matrix );

		void operator *= ( const long scalar ) {
			for ( uint_fast16_t index = 0; index < getRows(); index++ ) {
				mData[index] *= scalar;
			}
		}

		void operator *= ( const mpz_class& scalar ) {
			for ( uint_fast16_t index = 0; index < getRows(); index++ ) {
				mData[index] *= scalar;
			}
		}

		void operator *= ( const Matrix & rhs );

		Matrix operator * ( const Matrix & rhs ) const;

		RepresentationVector operator * ( const RepresentationVector* rhs ) const {

			RepresentationVector vector( rhs->getDimension() );

			for ( uint_fast16_t row = 0; row < getRows(); row++ ) {

				UtilsFactory::CALC_UTILS.loadRegisterS( 0 );

				for ( uint_fast16_t index = 0; index < vector.getDimension(); index++ ) {
					if ( rhs->isNonZeroAt( index ) && isNonZeroAt( row, index ) ) {
						UtilsFactory::CALC_UTILS.loadRegisterT( rhs->getValue( index ) );
						UtilsFactory::CALC_UTILS.multiplyRegisterT( getValue( row, index ) );

						UtilsFactory::CALC_UTILS.incrementRegisterS( UtilsFactory::CALC_UTILS.getRegisterT() );
					}
				}

				if ( UtilsFactory::CALC_UTILS.getRegisterS() != 0 ) {
					vector.setValue( row, UtilsFactory::CALC_UTILS.getRegisterS() );
				}
			}

			return vector;
		}

		Matrix getSubMatrix ( const uint_fast16_t firstRow, const uint_fast16_t lastRow ) const;

		void swapRows ( const uint_fast16_t first, const uint_fast16_t second ) {
			const uint_fast16_t swap_element = mMapping[ first ];
			mMapping[ first ] = mMapping[ second ];
			mMapping[ second ] = swap_element;
		}
		
		void sort ( const SortCriterion & );

		const std::string getMapping () const;
		const std::string toString () const;
        const std::string toPortaFormat () const;
        const std::string to4ti2Format () const;
        const std::string to4ti2RelFormat () const;
        const std::string to4ti2SignFormat () const;
	};
}

#endif // __DD_MATRIX__
