/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_OPTIONS__
#define __DD_OPTIONS__

#include <string>
#include <sstream>
#include <fstream>
#include <algorithm> 
#include <functional> 
#include <cctype>
#include <locale>

#include "DDLogging.h"

#define SEC_GENERAL_SETTINGS                "GENERAL"

#define SEC_NARROWING_LD_TREE_LD_INPUT      "NARROWING LD TREE FOR LD INPUT"
#define SEC_NARROWING_HD_TREE_LD_INPUT      "NARROWING HD TREE FOR LD INPUT"
#define SEC_NARROWING_HD_TREE_HD_INPUT      "NARROWING HD TREE FOR HD INPUT"

#define SEC_VERIFICATION_LD_TREE            "VERIFICATION LD TREE"
#define SEC_VERIFICATION_HD_TREE            "VERIFICATION HD TREE"

#define OPT_NARROWING_DEG_THRESHOLD         "Narrowing Degeneracy Threshold"
#define OPT_VERIFICATION_DEG_THRESHOLD      "Verification Degeneracy Threshold"

#define OPT_ADJACENCY_TEST                  "Adjacency Test"
#define OPT_PLOT_FILE                       "Plot file"
#define OPT_OUTPUT_FORMAT                   "Output format"
#define OPT_THREAD_EXPONENT                 "Thread Exponent"
#define OPT_INPUT_ORDERING                  "Input Ordering"

#define OPT_CONTAINER_TYPE                  "Active Set Container"
#define OPT_BIT_SELECT_STRATEGY             "Bit Selection Strategy"
#define OPT_FRAGMENTATION                   "Fragmentation"
#define OPT_MAX_TREE_DEPTH                  "Max Tree Depth"
#define OPT_LEAF_CAPACITY                   "Leaf Capacity"
#define OPT_DATA_SORT                       "Data Sort"

#define OPT_BITS_PER_SPLIT                  "Bits Per Split"
#define OPT_DYNAMIC_NARROWING_FACTOR        "DynamicNarrowingFactor"
#define OPT_DYNAMIC_BALANCE_FACTOR          "DynamicBalanceFactor"

#define OPT_VANTAGE_POINT_POPULATION        "VantagePointPopulation"
#define OPT_VANTAGE_POINT_DISTANCE_COEFF    "VantagePointDistanceCoefficient"
#define OPT_UNIQUE_VANTAGE_POINT            "UniqueVantagePoint"

namespace addibit {

    /* The adjacency test to be used. */
    enum AdjacencyTest {
        COMBINATORIAL_TEST,
        ALGEBRAIC_TEST,
        PARTITIONED_SETS_TEST
    };
    const std::string AdjacencyTestNames[] = {
        "COMBINATORIAL_TEST",
        "ALGEBRAIC_TEST",
        "PARTITIONED_SETS_TEST"
    };
    const uint_fast16_t AdjacencyTestNamesSize = 3;

    /* How to sort the rows of the input matrix? */
    enum InputOrdering {
        NO_ORDERING,
        LEX,
        COLEX,
        REV_LEX,
        REV_COLEX,
        MIN_ZERO,
        MAX_ZERO   
    };
    const std::string InputOrderingNames[] = {
        "NO_ORDERING",
        "LEX",
        "COLEX",
        "REV_LEX",
        "REV_COLEX",
        "MIN_ZERO",
        "MAX_ZERO"        
    };
    const uint_fast16_t InputOrderingNamesSize = 7;

    /* Which container type to use for the adjacency test? */
    enum TreeType {
        ARRAY,
        BIT_PATTERN_TREE,
        EXTENDED_BIT_PATTERN_TREE,
        FRAGMENTED_BIT_PATTERN_TREE,
        BINARY_DYNAMIC_VANTAGE_POINT,
        BINARY_STATIC_VANTAGE_POINT,
        TERNARY_DYNAMIC_VANTAGE_POINT,
        STATIC_POPULATION_VANTAGE_POINT,
        POPULATION_TREE,
        FRAGMENTATION_TREE
    };
    const std::string TreeTypeNames[] = {
        "ARRAY",
        "BIT_PATTERN_TREE",
        "EXTENDED_BIT_PATTERN_TREE",
        "FRAGMENTED_BIT_PATTERN_TREE",        
        "BINARY_DYNAMIC_VANTAGE_POINT",
        "VANTAGE_POINT_TREE",
        "TERNARY_DYNAMIC_VANTAGE_POINT",
        "STATIC_POPULATION_VANTAGE_POINT",
        "POPULATION_TREE",
        "FRAGMENTATION_TREE"
    };
    const uint_fast16_t TreeTypeNamesSize = 10;

    /* If a bit pattern tree is used, what partitioning strategy to employ? */
    enum BP_Tree_BitSelectionStrategy {
        BP_TREE_FIRST_AVAILABLE_BIT,
        BP_TREE_MOST_COMMON_BIT_FIRST,
        BP_TREE_LEAST_COMMON_BIT_FIRST,
        BP_TREE_MOST_AVERAGE_BIT_FIRST,
        BP_TREE_CLOSEST_TO_50_BIT_FIRST,
        BP_TREE_QUERY_BITS_NEUTRALIZATION,
        BP_TREE_INV_QUERY_BITS_NEUTRALIZATION,
        BP_TREE_LEAST_AVERAGE_BIT_FIRST,
        BP_TREE_GREATEST_POS_NEG_DIFF_FIRST
    };
    const std::string BP_Tree_BitSelectionStrategyNames[] = {
        "BP_TREE_FIRST_AVAILABLE_BIT",
        "BP_TREE_MOST_COMMON_BIT_FIRST",
        "BP_TREE_LEAST_COMMON_BIT_FIRST",
        "BP_TREE_MOST_AVERAGE_BIT_FIRST",
        "BP_TREE_CLOSEST_TO_50_BIT_FIRST",
        "BP_TREE_QUERY_BITS_NEUTRALIZATION",
        "BP_TREE_INV_QUERY_BITS_NEUTRALIZATION",
        "BP_TREE_LEAST_AVERAGE_BIT_FIRST",
        "BP_TREE_GREATEST_POS_NEG_DIFF_FIRST"
    };
    const uint_fast16_t BP_Tree_BitSelectionStrategyNamesSize = 9;

    /* How to output the results of the computation. */
    enum OutputFormat {
        INTEGER_VEC,
        DOUBLE_VEC,
        ACTIVE_SET,
        COUNT_ONLY
    };
    const std::string OutputFormatNames[] = {
        "INTEGER_VEC",
        "DOUBLE_VEC",
        "ACTIVE_SET",
        "COUNT_ONLY"
    };
    const uint_fast16_t OutputFormatNamesSize = 4;

    /* How to sort the active sets before starting the adjacency test? */
    enum SortType {
        MAX_SET_BITS,
        MIN_SET_BITS,
        MAX_SET_BITS_LEX,
        MIN_SET_BITS_LEX,
        MAX_BIT_FREQUENCY,
        MIN_BIT_FREQUENCY,
        MAX_SET_BITS_MAX_BIT_FREQUENCY,
        MAX_SET_BITS_MIN_BIT_FREQUENCY,
        MIN_SET_BITS_MAX_BIT_FREQUENCY,
        MIN_SET_BITS_MIN_BIT_FREQUENCY,
        NONE
    };
    const std::string SortTypeNames[] = {
        "MAX_SET_BITS",
        "MIN_SET_BITS",
        "MAX_SET_BITS_LEX",
        "MIN_SET_BITS_LEX",
        "MAX_BIT_FREQUENCY",
        "MIN_BIT_FREQUENCY",
        "MAX_SET_BITS_MAX_BIT_FREQUENCY",
        "MAX_SET_BITS_MIN_BIT_FREQUENCY",
        "MIN_SET_BITS_MAX_BIT_FREQUENCY",
        "MIN_SET_BITS_MIN_BIT_FREQUENCY",
        "NONE"
    };
    const uint_fast16_t SortTypeNamesSize = 11;

    class TreeOptions {

        private:
        const std::string mCaption;

        TreeType mTreeType;
        
        /* Only for bit pattern and population trees */
        BP_Tree_BitSelectionStrategy mBPT_Strategy;
        
        /* Only for bp-trees and fragmented bp-trees */
        uint_fast16_t mFBPT_Fragmentation;

        uint_fast16_t mMaxTreeDepth;
        uint_fast16_t mLeafCapacity;
        uint_fast16_t mMinElements;

        SortType mDataSort;

        uint_fast16_t mBitsPerSplit;
        double mDynamicNarrowingFactor;
        double mDynamicBalanceFactor;

        uint_fast16_t mVantagePointPopulation;
        double_t mVantagePointDistanceCoefficient;  
        bool mUniqueVantagePoint;

        public:
        TreeOptions ( const std::string caption ) :
            mCaption( caption ),
            mTreeType( BIT_PATTERN_TREE ),
            mBPT_Strategy( BP_TREE_FIRST_AVAILABLE_BIT ),
            mFBPT_Fragmentation( 0 ),
            mMaxTreeDepth( 15 ),
            mLeafCapacity( 20 ),
            mMinElements( 100 ),
            mDataSort( NONE ),
            mBitsPerSplit( 1 ),
            mDynamicNarrowingFactor( 1.5 ),
            mDynamicBalanceFactor( 0.0 ),
            mVantagePointPopulation( 50 ),
            mVantagePointDistanceCoefficient( 0.5 ),
            mUniqueVantagePoint( false ) {}  

        void set ( const std::string & name, const std::string & value );

        inline TreeType tree_type () const {
            return mTreeType;
        }

        inline BP_Tree_BitSelectionStrategy bpt_bit_selection_strategy () const {
            return mBPT_Strategy;
        }
        
        inline uint_fast16_t fbpt_fragmentation () const {
            return mFBPT_Fragmentation;
        }

        inline uint_fast16_t max_tree_depth () const {
            return mMaxTreeDepth;
        }

        inline uint_fast16_t leaf_capacity () const {
            return mLeafCapacity;
        }
        
        inline uint_fast16_t min_elements () const {
            return mMinElements;
        }

        inline uint_fast16_t bits_per_split () const {
            return mBitsPerSplit;
        }

        inline uint_fast16_t vantage_point_population () const {
            return mVantagePointPopulation;
        }

        inline double_t vantage_point_distance_coefficient () const {
            return mVantagePointDistanceCoefficient;
        }

        inline double dynamic_narrowing_factor () const {
            return mDynamicNarrowingFactor;
        }

        inline double dynamic_balance_factor () const {
            return mDynamicBalanceFactor;
        }

        inline SortType data_sort () const {
            return mDataSort;
        }         

        inline bool unique_vp () const {
            return mUniqueVantagePoint;
        }

        const std::string toString () const;
    };

    class Options {
        
        private:
        AdjacencyTest mAdjacencyTest;
        OutputFormat mOutputFormat;
        uint_fast16_t mThreadExponent;
        InputOrdering mInputOrdering;

        std::string mPlotFile;

        TreeOptions mNarrowingLDInputHDTree;
        TreeOptions mNarrowingLDInputLDTree;
        TreeOptions mNarrowingHDInputHDTree;
        uint_fast16_t mNarrowingDegeneracyThreshold;
        
        TreeOptions mVerificationLDTree;
        TreeOptions mVerificationHDTree;
        uint_fast16_t mVerificationDegeneracyThreshold;

        static const char COMMENTS_DELIM;
        static const char ASSIGNMENT_DELIM;

        Options () :
            mAdjacencyTest( COMBINATORIAL_TEST ),
            mOutputFormat( COUNT_ONLY ),
            mThreadExponent( 0 ),
            mInputOrdering( NO_ORDERING ),
            mPlotFile( "Don't plot" ),
            mNarrowingLDInputHDTree( SEC_NARROWING_HD_TREE_LD_INPUT ),
            mNarrowingLDInputLDTree( SEC_NARROWING_LD_TREE_LD_INPUT ),
            mNarrowingHDInputHDTree( SEC_NARROWING_HD_TREE_HD_INPUT ),
            mNarrowingDegeneracyThreshold( 0 ),
            mVerificationLDTree( SEC_VERIFICATION_LD_TREE ),
            mVerificationHDTree( SEC_VERIFICATION_HD_TREE ),
            mVerificationDegeneracyThreshold( 0 ) {}

        static Options * mCurrent;

        void set ( const std::string & name, const std::string & value );

        public:
        static Options & current () {

            if ( mCurrent == NULL ) {
                mCurrent = new Options();
            } 
            
            return *mCurrent;
        }        

        static void release () {
            if ( mCurrent != NULL ) {
                delete mCurrent;
            } 
        }

        static void trim ( std::string & s ) {
            s.erase( s.begin(), std::find_if(s.begin(), s.end(), std::not1(std::ptr_fun<int, int>(std::isspace))));
            s.erase( std::find_if(s.rbegin(), s.rend(), std::not1(std::ptr_fun<int, int>(std::isspace))).base(), s.end());
        }

        static void remove_comments ( std::string & s, const char delim ) {
            size_t p = s.find( delim );

            if ( p != std::string::npos ) {
                s.erase( s.begin() + p, s.end() );
            }
        }

        bool readSettingsFile ( const char * file );

        static void write_value ( const std::string & s_value, bool & b_value ) {
            std::istringstream is_value( s_value );
            is_value >> b_value;        
        }

        static void write_value ( const std::string & s_value, uint_fast16_t & ul_value ) {
            std::istringstream is_value( s_value );
            is_value >> ul_value;        
        }

        static void write_value ( const std::string & s_value, double & d_value ) {
            std::istringstream is_value( s_value );
            is_value >> d_value;        
        }

        template< typename EnumType >
        static void write_value ( const std::string & s_value,
                                  EnumType & e_value,
                                  const std::string* names,
                                  const uint_fast32_t size ) {
            const std::string * res = std::find( names, names + size, s_value );
            if ( res != names + size ) {
                e_value = static_cast<EnumType>( res - names );
            }   
        }

        OutputFormat output_format () const {
            return mOutputFormat;
        }

        AdjacencyTest adjacency_test () const {
            return mAdjacencyTest;
        }
   
        const TreeOptions & ld_narrowing () const {
            return mNarrowingLDInputLDTree;
        }   
        
        const TreeOptions & mixed_narrowing () const {
            return mNarrowingLDInputHDTree;
        } 
        
        const TreeOptions & hd_narrowing () const {
            return mNarrowingHDInputHDTree;
        }         
        
        inline uint_fast16_t narrowing_degeneracy_threshold () const {
            return mNarrowingDegeneracyThreshold;
        }  

        const TreeOptions & ld_verification () const {
            return mVerificationLDTree;
        }
        
        const TreeOptions & hd_verification () const {
            return mVerificationHDTree;
        }
        
        inline uint_fast16_t verification_degeneracy_threshold () const {
            return mVerificationDegeneracyThreshold;
        }
        
        uint_fast16_t thread_exponent () const {
            return mThreadExponent;
        }
        
        InputOrdering & input_ordering () {
            return mInputOrdering;
        }

        const std::string & plot_file () const {
            return mPlotFile;
        }

        bool plot_required () const {
            return mPlotFile != "Don't plot";
        }
    
        const std::string getCaptionBlock ( const std::string text ) const;
        const std::string toString () const;
    };
}

#endif // __DD_OPTIONS__
