/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_RAY_REFERENCE__
#define __DD_RAY_REFERENCE__

#include <math.h>

#include "DDBitList.h"
#include "DDNarrowingCandidate.h"
#include "DDEdgeCandidate.h"

namespace addibit {

	struct RayReference {
		BitList activeSet;
        unsigned int degeneracy;
        unsigned long frequency;

        uint_fast32_t id;
        uintptr_t __address;

		RayReference () {}

		RayReference ( const BitList & as ) :
			activeSet( as ),
            degeneracy( as.getPopulationCount() ),
            frequency( 0 ),
            id( 0 ),
            __address( 0 ) {}

		RayReference ( const RayReference & other, const uint_fast16_t _id ) :
			activeSet( other.activeSet ),
            degeneracy( other.degeneracy ),
            frequency( other.frequency ),
            id( _id ),
            __address( other.__address ) {}
            
        RayReference ( const BitList & as, const uint_fast16_t _id, const uintptr_t address ) :
			activeSet( as ),
            degeneracy( as.getPopulationCount() ),
            frequency( 0 ),
            id( _id ),
            __address( address ) {}
            
        RayReference ( BitList && as, const uint_fast16_t _id, const uintptr_t address ) :
			activeSet( std::move( as ) ),
            degeneracy( activeSet.getPopulationCount() ),
            frequency( 0 ),
            id( _id ),
            __address( address ) {}          
            
        RayReference ( const BitList & as, const uint_fast16_t _id, const uintptr_t address, const uint_fast16_t deg ) :
			activeSet( as ),
            degeneracy( deg ),
            frequency( 0 ),
            id( _id ),
            __address( address ) {} 
            
		RayReference ( RayReference && rvalue ) :
			activeSet( rvalue.activeSet ),
            degeneracy( rvalue.degeneracy ),
            frequency( rvalue.frequency ),
            id( rvalue.id ),
            __address( rvalue.__address ) {}                   
        
        void operator= ( RayReference && rvalue ) {
            activeSet = std::move( rvalue.activeSet );
            degeneracy = rvalue.degeneracy;
            id = rvalue.id;
            __address = rvalue.__address;
        } 
		
        const BitList & operator() () const {
		    return activeSet;
		}
	};

    class FragmentedRayReference : public NarrowingCandidate {
        
        private:
        const BitList mFragMap;

        public:
        FragmentedRayReference ( const NarrowingCandidate & cand,
                                 const BitList & fmap ) :
			NarrowingCandidate( cand ),
            mFragMap( fmap ) {}
         
		const BitList & operator() () const {
		    return mFragMap;
		}        
    };
        
        
    class FragmentedEdgeCandidate : public EdgeCandidate {
    
        private:
        const BitList mFragMap;
        
        public:
        FragmentedEdgeCandidate ( const EdgeCandidate & cand,
                                  const BitList & fmap ) :
            EdgeCandidate( cand ),
            mFragMap( fmap ) {}
    
		const BitList & operator() () const {
		    return mFragMap;
		}        
    };

    struct RefComparatorMostBitsFirst {

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy > rhs.degeneracy;
        }
    };

    struct RefComparatorLeastBitsFirst {

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy < rhs.degeneracy;
        }
    };

    struct RefComparatorMostBitsFirstLex {

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy > rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.activeSet.lex_less( rhs.activeSet ) );
        }
    };

    struct RefComparatorLeastBitsFirstLex {

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy < rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.activeSet.lex_less( rhs.activeSet ) );
        }
    };

    struct RefComparatorClosestToPoint {

        const BitList & mPoint;

        RefComparatorClosestToPoint ( const BitList & point ) :
            mPoint( point ) {}

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            const unsigned int lhs_dist = mPoint.getCommonPopCnt( lhs.activeSet );
            const unsigned int rhs_dist = mPoint.getCommonPopCnt( rhs.activeSet );

            return lhs_dist < rhs_dist || ( lhs_dist == rhs_dist && mPoint.getInvCommonPopCnt( lhs.activeSet ) < mPoint.getInvCommonPopCnt( rhs.activeSet ) ) ;
        }        
    };

    struct RefComparatorClosestToMultiplePoints {

        const BitList * mPoints;
        const unsigned int mCount;

        RefComparatorClosestToMultiplePoints ( const BitList * points, const unsigned int count ) :
            mPoints( points ),
            mCount( count ) {}

        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            unsigned int lhs_dist;
            unsigned int rhs_dist;

            unsigned int index = 0;

            do {
                lhs_dist = mPoints[ index ].getCommonPopCnt( lhs.activeSet );
                rhs_dist = mPoints[ index ].getCommonPopCnt( rhs.activeSet );

                index++;
            } while ( lhs_dist == rhs_dist && index < mCount );

            return lhs_dist < rhs_dist;
        }        
    };

    struct RefComparatorMaxFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.frequency > rhs.frequency;    
        }
    };

    struct RefComparatorMinFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.frequency < rhs.frequency;    
        }
    };

    struct RefComparatorMaxSetBitsMaxFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy > rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.frequency > rhs.frequency );    
        }
    };

    struct RefComparatorMaxSetBitsMinFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy > rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.frequency < rhs.frequency );    
        }
    };

    struct RefComparatorMinSetBitsMinFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy < rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.frequency < rhs.frequency );    
        }
    };

    struct RefComparatorMinSetBitsMaxFrequency {
    
        bool operator() ( const struct RayReference & lhs, const struct RayReference & rhs ) const {
            return lhs.degeneracy < rhs.degeneracy || ( lhs.degeneracy == rhs.degeneracy && lhs.frequency > rhs.frequency );    
        }
    };
}

#endif // __DD_RAY_REFERENCE__
