/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_RAY_REFERENCE_CONTAINER__
#define __DD_RAY_REFERENCE_CONTAINER__

#include <algorithm>

#include "DDTweaks.h"

#include "DDInterval.h"
#include "DDContainerBucket.h"
#include "DDRayRefVisitor.h"
#include "DDAdjacencyTestStack.h"

#define __DD_BUCKET_CAPACITY__  64
#define __DD_BUCKET_ALLOC_SPACE__ new RayReference[ __DD_BUCKET_CAPACITY__ ]

namespace addibit {

    class EmptyTransformation {
        
        public:        
        void apply ( const NarrowingCandidate *, AdjacencyTestStack * ) const {}  
        void apply ( EdgeCandidate & ) const {}
        
        const EdgeCandidate & execute ( const EdgeCandidate & candidate ) const {
            return candidate;
        }                
        
        uint_fast16_t getNarrowingCandidateSize () const {
            return sizeof( NarrowingCandidate ) + sizeof( __BIT_LIST_RAW_DATA_TYPE__ ) * ( __DD_BLOCK_COUNT__ - 1 );
        }
        
        uint_fast16_t getAdjacencyCandidateSize () const {
            return sizeof( AdjacencyCandidate ) + sizeof( __BIT_LIST_RAW_DATA_TYPE__ ) * ( __DD_BLOCK_COUNT__ - 1 );
        }        
        
        uint_fast16_t getAdjacencyStackSize () const {
            return 0;
        }
    };

    class RayReferenceContainer {

        private:
        uint_fast32_t mId;        
        uint_fast32_t mCapacity;
        
        inline void trim () {
            mCapacity -= __mem_region_end - __next_element_ptr;
            __mem_region_end = __next_element_ptr;
        }        
        
        protected:
        uint_fast16_t mBuckets;
        ContainerBucket * __bucket_array;
        
        RayReference * __next_element_ptr;
        RayReference * __mem_region_end;        
        
        ContainerBucket & create_new_bucket () {
            ContainerBucket * __old_bucket_array = __bucket_array;
            __bucket_array = new ContainerBucket[ mBuckets + 1 ];
            
            for ( uint_fast16_t b_index = 0; b_index < mBuckets; b_index++ ) {
                __bucket_array[ b_index ] = std::move( __old_bucket_array[ b_index ] );
            }
            __bucket_array[ mBuckets ] = ContainerBucket( __DD_BUCKET_ALLOC_SPACE__, __DD_BUCKET_CAPACITY__ );
            delete[] __old_bucket_array;
            
            return __bucket_array[ mBuckets++ ];
        }
        
        inline void ensure_enough_capacity () {
            if ( unlikely( full() ) ) {
                ContainerBucket & new_bucket = create_new_bucket();
                __next_element_ptr = new_bucket.start_ptr();
                __mem_region_end = new_bucket.end_ptr();
                
                mCapacity += new_bucket.capacity();
            }
        }
        
        template<class VISITOR>
        inline void run_init_round ( VISITOR & visitor ) const {
            
            if ( visitor.needsInitRound() ) {
                visitor.onInitStart();
                for ( uint_fast16_t b_index = 0; b_index < mBuckets - 1; b_index++ ) {
                    __bucket_array[ b_index ].iterate_init( visitor );
                }
                __bucket_array[ mBuckets - 1 ].iterate_init( visitor, __next_element_ptr );            
            }
        }            
        
        inline RayReference & first () {
            return __bucket_array[0].start_ptr()[0];
        }

        inline RayReference & last () {
            return ( __next_element_ptr - 1 )[0];
        }        
        
        public:

        RayReferenceContainer ( RayReference * mem = __DD_BUCKET_ALLOC_SPACE__,
                                const uint_fast32_t size = __DD_BUCKET_CAPACITY__ ) :
            mId( 0 ),
            mBuckets( 1 ),
            __bucket_array( new ContainerBucket[ 1 ] ) {
                
            __bucket_array[0] = ContainerBucket( mem, size );
            __next_element_ptr = __bucket_array[0].start_ptr();
            __mem_region_end = __bucket_array[0].end_ptr();
            
            mCapacity = __bucket_array[0].capacity();
        }
        
        RayReferenceContainer ( ContainerBucket * bucket_array,
                                const uint_fast16_t buckets,
                                const uint_fast32_t capacity ) :
            mId( 0 ),
            mCapacity( capacity ),
            mBuckets( buckets ),
            __bucket_array( bucket_array ),
            __next_element_ptr( bucket_array[ buckets - 1 ].end_ptr() ),
            __mem_region_end( __next_element_ptr ) {}
        
        RayReferenceContainer ( RayReferenceContainer && rvalue ) :
            mId( rvalue.mId ),
            mBuckets( rvalue.mBuckets ),
            __bucket_array( rvalue.__bucket_array ),
            __next_element_ptr( rvalue.__next_element_ptr ),
            __mem_region_end( rvalue.__mem_region_end ) {
        
            rvalue.__bucket_array = NULL;
            rvalue.__next_element_ptr = NULL;
            rvalue.__mem_region_end = NULL;
        }
        
        ~RayReferenceContainer () {
            destroy();
        }
        
        /* Precondition: full() == true && rvalue.full() == true */
        RayReferenceContainer operator+ ( RayReferenceContainer && rvalue ) {
            
            ContainerBucket * bucket_array = new ContainerBucket[ mBuckets + rvalue.mBuckets ];
            
            for ( uint_fast16_t b_index = 0; b_index < mBuckets; b_index++ ) {
                bucket_array[ b_index ] = std::move( __bucket_array[ b_index ] );
            }
            
            for ( uint_fast16_t b_index = 0; b_index < rvalue.mBuckets; b_index++ ) {
                bucket_array[ b_index + mBuckets ] = std::move( rvalue.__bucket_array[ b_index ] );
            } 
            
            return RayReferenceContainer( bucket_array, mBuckets + rvalue.mBuckets, capacity() + rvalue.capacity() );
        }       

        inline bool full () const {
            return __next_element_ptr == __mem_region_end;
        }

        inline uint_fast32_t id () const {
            return mId;
        }
        
        inline void setId ( const uint_fast32_t id ) {
            mId = id;
        }
        
        inline void insert ( RayReference && ref ) {
            ensure_enough_capacity();
            ( __next_element_ptr++ )[ 0 ] = std::move( ref );
        }
        
        inline void insert ( BitList && as, const uint_fast16_t id, const uintptr_t address ) {
            ensure_enough_capacity();
            ( __next_element_ptr++ )[ 0 ] = RayReference( std::move( as ), id, address );
        }     

        inline uint_fast32_t size () const {
            return mCapacity - ( __mem_region_end - __next_element_ptr );
        }
        
        inline uint_fast32_t capacity () const {
            return mCapacity;
        }

        inline bool empty () const {
            return mBuckets == 1 && __next_element_ptr == __bucket_array[0].start_ptr();
        }

        inline void rotate () {
            std::swap( first(), last() );
        }

        ContainerBucket & bucket ( const uint_fast16_t i ) {
            return __bucket_array[i];
        }

        const RayReference & first () const {
            return __bucket_array[0].start_ptr()[0];
        }

        const RayReference & last () const {
            return ( __next_element_ptr - 1 )[0];
        }

        const RayReference & at ( const uint_fast32_t pos ) const {
            return __bucket_array[ pos / __DD_BUCKET_CAPACITY__ ].start_ptr()[ pos % __DD_BUCKET_CAPACITY__ ];
        }

        const RayReference & operator[] ( const uint_fast32_t pos ) const {
            return at( pos );
        }
  
        template<class VISITOR> 
        void iterate ( VISITOR & visitor ) const {  

            run_init_round( visitor );

            visitor.onStart();
            for ( uint_fast16_t b_index = 0; b_index < mBuckets - 1; b_index++ ) {
                __bucket_array[ b_index ].iterate( visitor );
            }
            __bucket_array[ mBuckets - 1 ].iterate( visitor, __next_element_ptr );  
            visitor.onFinish();
        }       
        
        template<class VISITOR> 
        void successive_pop ( VISITOR & visitor ) {  

            run_init_round( visitor );

            visitor.onStart();
            for ( uint_fast16_t b_index = 0; b_index < mBuckets - 1; b_index++ ) {
                __bucket_array[ b_index ].successive_pop( visitor );
            }
            __bucket_array[ mBuckets - 1 ].successive_pop( visitor, __next_element_ptr );  
            visitor.onFinish();
        }

        void modify ( const RayRefModifier & modifier ) {
            for ( uint_fast16_t b_index = 0; b_index < mBuckets - 1; b_index++ ) {
                __bucket_array[ b_index ].modify( modifier );
            }
            __bucket_array[ mBuckets - 1 ].modify( modifier, __next_element_ptr ); 
        }

       template<class Comparator> void sort ( const Comparator & comp ) {
         // @FIXME std::sort(  mRefs.begin(), mRefs.end(), comp );
        }
       
        void destroy () {
            if ( __bucket_array ) {
                delete[] __bucket_array;
                __bucket_array = NULL;
            }
        }
        
        const std::string toString () const;
    };
}

#endif // __DD_RAY_REFERENCE_CONTAINER__
