/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_SLE_SOLVER__
#define __DD_SLE_SOLVER__

#include "DDCalculationMatrix.h"
#include "DDLinearDependenceEntry.h"

#include "DDLogging.h"

namespace addibit {

	class SLESolver : public CalculationMatrix {

		private:
		uint_fast16_t mLastRowIndex;

		BitList mRowMap;

		protected:
        
        void prepare_for_output ( const unsigned int beginColumn ) {

            UtilsFactory::CALC_UTILS.startIncrementalLCM ( getValue( 0, 0 ) );        

            const unsigned int rows = getRows();

            for ( unsigned int row = 1; row < rows; row++ ) {
                UtilsFactory::CALC_UTILS.calcIncrementalLCM ( getValue( row, row ) );
            }

			for ( unsigned int index = 0; likely( index < rows ); index++ ) {
                CalculationVector & current_row = getRow( index );

                if ( current_row.isNonZeroAt( index ) ) {
                    UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.getResult() );
                    UtilsFactory::CALC_UTILS.divideRegisterS( current_row[ index ] );                

                    current_row.multiply( beginColumn, getColumns(), UtilsFactory::CALC_UTILS.get_register_s() );
                    current_row[ index ] = UtilsFactory::CALC_UTILS.getResult();
                }
			} 
        }

        void cut_zero_rows () {
			setRows( getColumns() - 1 );

			for ( int row = getRows() - 1; row >= 0 && !isNonZeroAt( row, row ); row-- ) {
				setColumns( getColumns() - 1 );
				setRows( getRows() - 1 );
			} 
        }

		public:
		SLESolver ( const uint_fast16_t rows, const uint_fast16_t columns );
		~SLESolver () {}

		void load ( const Matrix& matrix, const BitList& rowMap ) {

			unsigned int last_column = 0;

			for ( uint_fast16_t block = 0, prefix = 0; block < GLOBAL_BLOCKS; block++, prefix += __BLOCK_SIZE__ ) {
				unsigned long block_data = rowMap[ block ];

				while ( block_data ) {
					int suffix = __builtin_ctzl( block_data ); 
					unsigned int r_index= prefix + suffix;

                    const CalculationVector & row = matrix.getRow( r_index );
                    std::vector<mpz_class>::const_iterator r_it = row.iterator( 0 );
                    const std::vector<mpz_class>::const_iterator r_it_end = row.iterator( matrix.getColumns() );

					for ( unsigned int column = 0; likely( r_it != r_it_end ); column++, r_it++ ) {
						setValue( column, last_column, *r_it );
					}
					last_column++;

					block_data ^= ( __SINGLE_BIT__ << suffix );
				}
			}

			mRowMap = rowMap;
			setRows( matrix.getColumns() );
			setColumns( last_column );
		}

		void load ( const Matrix & matrix, const unsigned int target ) {

			unsigned int last_column = getColumns();

            const CalculationVector & target_row = matrix.getRow( target );
            std::vector<mpz_class>::const_iterator r_it = target_row.iterator( 0 );
            const std::vector<mpz_class>::const_iterator r_it_end = target_row.iterator( matrix.getColumns() );

			for ( unsigned int column = 0; likely( r_it != r_it_end ); column++, r_it++ ) {
				setValue( column, last_column, *r_it );
			}

			mLastRowIndex = target;
			setColumns( last_column + 1 );
		}
		
		void load ( const Matrix & matrix, const BitList & rowMap, const unsigned int target ) {

			unsigned int last_column;

			load( matrix, rowMap );			

			last_column = getColumns();

			for ( unsigned int column = 0; column < matrix.getColumns(); column++ ) {
				setValue( column, last_column, matrix[target][column] );
			}

			mLastRowIndex = target;
			setColumns( last_column + 1 );
		}

		void solve () {
			transformToUpperTriangularMatrix();
			transformToDiagonalMatrix();
			prepare_for_output( getRows() - 1 ); 
		}

		void fillLinearDependenceEntry ( LinearDependenceEntry * const entry ) {
			const uint_fast16_t last_column = getColumns() - 1;

            const mpz_class & start_value = getValue( 0, 0 );
			entry->setScalingFactor( start_value );
            UtilsFactory::CALC_UTILS.startIncrementalGCD( start_value );

			for ( uint_fast16_t block = 0, prefix = 0, row = 0; block < GLOBAL_BLOCKS; block++, prefix += __BLOCK_SIZE__ ) {
				unsigned long block_data = mRowMap[block];

				while ( block_data ) {
					const uint_fast16_t suffix = __builtin_ctzl( block_data ); 
					const uint_fast16_t index = prefix + suffix;

					if ( getRow( row ).isNonZeroAt( last_column ) ) {
                        const mpz_class & current_value = getValue( row, last_column );
						entry->insert( index, current_value );
                        UtilsFactory::CALC_UTILS.calcIncrementalGCD( current_value );
					}
					row++;

					block_data ^= ( __SINGLE_BIT__ << suffix );
				}
			}

            *entry /= UtilsFactory::CALC_UTILS.getResult();

			setColumns( last_column );
		}

		const LinearDependenceEntry * getLinearDependenceEntry () const {

			LinearDependenceEntry * result = new LinearDependenceEntry( mLastRowIndex + 1, getValue( 0, 0 ) );
			const unsigned int last_column = getColumns() - 1;

			for ( uint_fast16_t block = 0, prefix = 0, r_index = 0; block < GLOBAL_BLOCKS; block++, prefix += __BLOCK_SIZE__ ) {
				unsigned long block_data = mRowMap[block ];

				while ( block_data ) {
					int suffix = __builtin_ctzl( block_data ); 
					unsigned int index = prefix + suffix;

                    const CalculationVector & row = getRow( r_index );

					if ( row.isNonZeroAt( last_column ) ) {
						result->insert( index, row[ last_column ] );
					}
					r_index++;

					block_data ^= ( __SINGLE_BIT__ << suffix );
				}
			}

			return result;
		}
	};
}

#endif // __DD_SLE_SOLVER__
