/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_STATIC_BIT_LIST__
#define __DD_STATIC_BIT_LIST__

namespace addibit {

	template<unsigned int BLOCKS>
	class StaticBitList {

		private:
		unsigned long data[ BLOCKS ];  

		inline void init_data ( const unsigned long value ) {
			for ( unsigned int block = 0;  block < BLOCKS; block++ ) {
				data[ block ] = value;
			}
		}

		public:

		StaticBitList () {
 			init_data( 0 );
		}

		StaticBitList ( const unsigned long value )  {
			init_data( value );
		}

		StaticBitList ( const StaticBitList<BLOCKS> & other ) {
			for ( unsigned int block = 0;  block < BLOCKS; block++ ) {
				data[ block ] = other.data[ block ];
			} 
		}

		~StaticBitList () {}

		static const BitCountTable BIT_COUNT_TABLE;

		static unsigned long getBlocks ( const unsigned long capacity ) {
			unsigned long block_count = capacity / __BLOCK_SIZE__ + (( capacity % __BLOCK_SIZE__ == 0 ) ? 0 : 1 );

			return block_count;
		}

		inline unsigned int getBlocks () const {
			return BLOCKS;
		}

		inline unsigned long getCapacity () const {
			return getBlocks() * __BLOCK_SIZE__;
		}

		bool isEmpty () const {
			bool result = ( getData( 0 ) == 0 );

			for ( unsigned int block = 1; block < BLOCKS && result; block++ ) {
				if ( getData( block ) != 0 ) {
					return false;
				}
			}

			return result;
		}

		bool isFull () const {
			bool result = ( getData( 0 ) == ULONG_MAX );

			for ( unsigned int block = 1; block < BLOCKS && result; block++ ) {
				if ( getData( block ) != ULONG_MAX ) {
					result = false;
					break;
				}
			}

			return result;
		}

		bool isUnionFull ( const StaticBitList<BLOCKS> & other ) const {
			bool result = ( (getData( 0 ) | other.getData( 0 )) == ULONG_MAX );

			for ( unsigned long block = 1; block < BLOCKS && result; block++ ) {
				if ( (getData( block ) | other.getData( block )) != ULONG_MAX ) {
					result = false;
					break;
				}
			}

			return result;
		}

		inline const unsigned long & getData ( const unsigned int block ) const {
			return data[ block ];
		}

		inline unsigned long & getData ( const unsigned int block ) {
			return data[ block ];
		}

		const StaticBitList<BLOCKS> operator& ( const StaticBitList<BLOCKS> & rhs ) const {
			StaticBitList<BLOCKS> result;

			for ( unsigned int block = 0; block < BLOCKS; block++ ) {
				result[ block ] = getData( block ) & rhs[ block ];
			}

			return result;
		}

		const StaticBitList<BLOCKS> operator| ( const StaticBitList<BLOCKS> & rhs ) const {
			StaticBitList<BLOCKS> result;

			for ( unsigned int block = 0; block < BLOCKS; block++ ) {
				result[ block ] = getData( block ) | rhs[ block ];
			}

			return result;
		}

		const StaticBitList<BLOCKS> operator^ ( const StaticBitList<BLOCKS> & rhs ) const {
			StaticBitList<BLOCKS> result;

            result[ 0 ] = data[ 0 ] ^ rhs[ 0 ];

			for ( unsigned int block = 0; block < BLOCKS; block++ ) {
				result[ block ] = getData( block ) ^ rhs.getData( block );
			}

			return result;
		}

		const StaticBitList<BLOCKS> operator~ () const {
			StaticBitList<BLOCKS> result;

			for ( unsigned int block = 0; block < BLOCKS; block++ ) {
				result[ block ] = ~getData( block );
			}

			return result;
		}

        const StaticBitList<BLOCKS> operator- ( const StaticBitList<BLOCKS> & rhs ) const {
            return (*this) ^ ( (*this) & rhs );
        }

		bool operator== ( const StaticBitList<BLOCKS> & rhs ) const {

			for ( unsigned int block = 0; block < BLOCKS; block++ ) {
				if ( getData( block ) != rhs[ block ] ) {
					return false;
			    }
            } 

			return true;
		}

		bool operator!= ( const  StaticBitList<BLOCKS> & rhs ) const {
	
			return !operator==( rhs );
		}

        bool lex_less ( const StaticBitList<BLOCKS> & rhs ) const {
            unsigned int block = 0;                  
      
			for ( ; block + 1 < BLOCKS && getData( block ) == rhs.getData( block ); block++ );

            return getData( block ) < rhs.getData( block );
        }
        
		void operator ^= ( const StaticBitList<BLOCKS> & rhs ) {

			for ( unsigned int block = 0; block < BLOCKS; block++ ) {
				data[ block ] ^= rhs[ block ];
			}
		}    
		
		/* Precondition: shift < 64! */
		void operator <<= ( const unsigned int shift ) {

            unsigned long transfer = data[ 0 ] >> ( __BLOCK_SIZE__ - shift );
			data[ 0 ] <<= shift;			

			for ( unsigned int block = 1; block < BLOCKS; block++ ) {
			    const unsigned long buffer = data[ block ] >> ( __BLOCK_SIZE__ - shift );
				data[ block ] = ( data[ block ] << shift ) | transfer;
				transfer = buffer;
			}
		}     		    

		void operator &= ( const StaticBitList<BLOCKS> & rhs ) {

			for ( unsigned int block = 0; block < BLOCKS; block++ ) {
				data[ block ] &= rhs[ block ];
			}
		}

		void operator |= ( const StaticBitList<BLOCKS> & rhs ) {		

			for ( unsigned int block = 0; block < BLOCKS; block++ ) {
				data[ block ] |= rhs[ block ];
			}
		}

		void operator= ( const StaticBitList<BLOCKS> & rhs ) {
		
			for ( unsigned int block = 0; block < BLOCKS; block++ ) {
				data[ block ] = rhs[ block ];
            }
		}

		const unsigned long & operator[] ( const unsigned int index ) const {
			return getData( index );
		}
		unsigned long& operator[] ( const unsigned int index ) {
			return getData( index );
		}

		void clear () {		
			init_data( 0 );
		}

		unsigned int getPopulationCount () const {
			unsigned int count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] );			

			for ( unsigned int block = 1; block < BLOCKS; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) );
			}

			return count;
		}

		unsigned int getInvPopulationCount () const {
			unsigned int count = BitList::BIT_COUNT_TABLE.lookUp ( ~data[0] );			

			for ( unsigned int block = 1;  block < BLOCKS; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( ~getData( block ) );
			}

			return count;
		}

        unsigned int getHammingDistance ( const StaticBitList<BLOCKS> & other ) const {
			unsigned int count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] xor other.data[0] );			

			for ( unsigned int block = 1;  block < BLOCKS; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) xor other.getData( block ) );
			}

			return count;
        }
        
        unsigned long getReducedMap ( const unsigned long map, const unsigned long red ) const {
            return (( map & red ) >> 1 ) | map; 
        }

		unsigned int getReducedCommonPopCnt ( const StaticBitList<BLOCKS> & other, const StaticBitList<BLOCKS> & reduction ) const {
			unsigned int count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] & getReducedMap( other.data[0], reduction[0] ) );			

			for ( unsigned int block = 1;  block < BLOCKS; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) & getReducedMap( other.getData( block ), reduction.getData( block ) ));
			}

			return count;		
		}

		unsigned int getCommonPopCnt ( const StaticBitList<BLOCKS> & other ) const {
			unsigned int count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] & other.data[0] );			

			for ( unsigned int block = 1;  block < BLOCKS; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) & other.getData( block ) );
			}

			return count;
		}
		
		unsigned int getDiffPopCnt ( const StaticBitList<BLOCKS> & other ) const {
			unsigned int count = BitList::BIT_COUNT_TABLE.lookUp ( ( data[0] ^ other.data[0] ) & data[0] );			

			for ( unsigned int block = 1;  block < BLOCKS; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( ( getData( block ) ^ other.getData( block ) ) & getData( block ) );
			}

			return count;
		}		

		unsigned int getCommonPopCnt ( const unsigned long * other_data ) const {
			unsigned int count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] & other_data[0] );			

			for ( unsigned int block = 1;  block < BLOCKS; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) & other_data[ block ] );
			}

			return count;
		}

		unsigned int getCommonPopCnt ( const StaticBitList<BLOCKS> & other, unsigned int & inv_pop_cnt ) const {
			unsigned int count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] & other.data[0] );	
            inv_pop_cnt = BitList::BIT_COUNT_TABLE.lookUp ( ~data[0] & other.data[0] );		

			for ( unsigned int block = 1;  block < BLOCKS; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) & other.getData( block ) );
                inv_pop_cnt += BitList::BIT_COUNT_TABLE.lookUp ( ~getData( block ) & other.getData( block ) );
			}

			return count;
		}

		unsigned int getInvCommonPopCnt ( const StaticBitList<BLOCKS> & other ) const {
			unsigned int count = BitList::BIT_COUNT_TABLE.lookUp ( ~data[0] & other.data[0] );			

			for ( unsigned int block = 1;  block < BLOCKS; block++ ) {
				count += BitList::BIT_COUNT_TABLE.lookUp ( ~getData( block ) & other.getData( block ) );
			}

			return count;
		}

		unsigned int getFirstSet () const {
			unsigned int block = 0;

			for ( block = 0; unlikely( block + 1 < BLOCKS && !getData(block) ); block++ );

			return block * __BLOCK_SIZE__ + __builtin_ctzl( getData( block ));
		}

		unsigned int getFirstUnset () const {
			unsigned long block = 0;

			for ( block = 0; unlikely( block + 1 < BLOCKS && !(~getData( block ))); block++ );

			return block * __BLOCK_SIZE__ + __builtin_ctzl( ~getData( block ));
		}

        unsigned int getNextSet ( unsigned int start ) const {
            unsigned long block = start / __BLOCK_SIZE__;
            unsigned long offset = start % __BLOCK_SIZE__;
            unsigned int result;

            register unsigned long pattern = ( ULONG_MAX << offset ) & getData( block );

            if ( __builtin_expect( pattern != 0, 1 ) ) {
                result = block * __BLOCK_SIZE__ + __builtin_ctzl( pattern );
            } else if ( __builtin_expect( block + 1 >= BLOCKS, 0 ) ) {
                result = getCapacity();
            } else {
                for ( block++ ; unlikely( block + 1 < BLOCKS && getData( block ) != 0 ); block++ );

                if ( getData( block ) != 0 ) {
                    result = block * __BLOCK_SIZE__ + __builtin_ctzl( getData( block ));
                } else {
                    result = getCapacity();
                }
            }

            return result;
        }

		bool isSet ( const unsigned long position ) const {
			return ( getData( position / __BLOCK_SIZE__ ) & ( __SINGLE_BIT__ << ( position % __BLOCK_SIZE__ ) ) );
		}

		void setBit ( const unsigned long position ) {
			getData( position / __BLOCK_SIZE__ ) |= ( __SINGLE_BIT__ << ( position % __BLOCK_SIZE__ ) );
		}

		void opOR ( const unsigned long block, const unsigned long mask ) {
			getData( block ) |= mask;
		}

		void setBits ( const unsigned long begin, const unsigned long end ) {

			for ( unsigned long index = begin; index < end; index++ ) {
				setBit( index );
			}
		}

        void set ( const unsigned long pattern ) {

            data[ 0 ] = pattern;

            for ( unsigned int block = 1;  block < BLOCKS; block++ ) {
			    data[ block ] = pattern;
            }        
        }

		void unsetBits ( const unsigned long begin, const unsigned long end ) {
			for ( unsigned long index = begin; index < end; index++ ) {
				unsetBit( index );
			}			
		}

		void unsetBitsFrom ( const unsigned long begin ) {
			for ( unsigned long index = begin; index < getCapacity() ; index++ ) {
				unsetBit( index );
			}			
		}

		void unsetBit ( const unsigned long position ) {	
			getData( position/ __BLOCK_SIZE__ ) &= ~( __SINGLE_BIT__ << ( position % __BLOCK_SIZE__ ) );
		}

		void opXOR ( const unsigned long block, const unsigned long mask ) {
			getData( block ) ^= mask;
		}

		bool containsUnexpected ( const unsigned long * const map ) const {
			bool result = ( map[0] & data[0] ) == map[0];

			for ( unsigned int block = 1;  block < BLOCKS && unlikely( result ); block++ ) {
				result = ( map[block] & getData( block ) ) == map[block];		
			} 

			return result;
		}

		bool contains ( const StaticBitList<BLOCKS> & map ) const {
			bool result = ( map[0] & data[0] ) == map[0];

			for ( unsigned int block = 1;  block < BLOCKS && result; block++ ) {
				result = ( map[block] & getData( block ) ) == map[block];
			} 

			return result;
		}
		
		unsigned long getReducedVerificationMap ( const unsigned long map_block, const unsigned long red_block ) const {
		    return map_block & ( ~red_block | ( red_block & ( map_block << 1 )));
		}
		
		bool contains ( const StaticBitList<BLOCKS> & map, const StaticBitList<BLOCKS> & reduction ) const {

			bool result = ( data[0] | getReducedVerificationMap( map[0], reduction[0] ) ) == data[0];
	
			for ( unsigned int block = 1;  block < BLOCKS && result; block++ ) {
				result = ( getData( block ) | getReducedVerificationMap( map.getData( block ), reduction.getData( block ) )) == getData( block );
			} 

			return result;
		}		
	};
}

#endif /* __DD_STATIC_BIT_LIST__ */
