/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_TREE_CREATOR__
#define __DD_TREE_CREATOR__

#include <utility>
#include <stdlib.h>

#include "DDAbstractTreeCreator.h"
#include "DDConstants.h"

#include "DDRootTreeNode.h"
#include "DDLeafNode.h"
#ifdef __DD_RECURSIVE_QUERY__
#include "DDIntermediateTreeNode.h"
#endif
#include "DDConditionedTreeNode.h"

#include "DDBitSplitVisitor.h"
#include "DDRayReferenceContainer.h"
#include "DDConditionedRayReferenceContainer.h"
#include "DDConditionedRayReferenceContainerList.h"
#include "DDConditionedTreeNode.h"

namespace addibit {

    template<class HD_COND_TYPE, class LD_COND_TYPE, class SPLITTER, class TRANSFORMATOR>
    class TreeCreator : public AbstractTreeCreator {
    
        private:
    #ifndef __DD_RECURSIVE_QUERY__
        char * __memory;
        char ** __memory_regions;
        uintptr_t * __ref_array;
        uintptr_t ** __ref_array_regions;
        uint_fast16_t __memory_regions_count;
        
        uint_fast16_t __cond_data_segment_approx_size;
        
        inline char * __alloc ( const uint_fast32_t size, const uint_fast16_t back_alignment = 8 ) {
            const uint_fast16_t offset = (uintptr_t)( __memory + size ) % back_alignment;
            char * __address = __memory + offset;
            __memory += size + offset;
            return __address;
        } 
        
        inline char * __sim_alloc ( const uint_fast32_t size, const uint_fast16_t back_alignment = 8 ) const {
            return __memory + (uintptr_t)( __memory + size ) % back_alignment;
        }
    #endif
        
        SPLITTER * const mVisitor;
        const TRANSFORMATOR * const mTransformator;
        const uint_fast32_t mMinElements;
        
#ifndef __DD_RECURSIVE_QUERY__
        template<class CONDITION_TYPE>
		inline ConditionedTreeNode * createChildNode ( ConditionedTreeNode * const ,
                                                       RayReferenceContainer & candidates,
                                                       CONDITION_TYPE && cond,                            
                                                       const bool leaf ) {
            ConditionedTreeNode * result;
            char * __cond_data_segment = __alloc( sizeof( ConditionedTreeNode ) + cond.mem_size() - sizeof( __BIT_LIST_RAW_DATA_TYPE__ ), CONDITION_TYPE::required_alignment() );

	        if ( leaf ) {
                char * __maps_data_segment = __alloc( candidates.size() * __DD_BLOCK_COUNT__ * sizeof( __BIT_LIST_RAW_DATA_TYPE__ ) );
	            result = new( __cond_data_segment ) ConditionedTreeNode( candidates, __maps_data_segment, __ref_array );
                __ref_array += candidates.size();
	        } else {        
                result = new( __cond_data_segment ) ConditionedTreeNode();
	        }    
            
            if ( build_for_narrowing() ) {
                cond.write_narrowing_data( result->data_segment() );
            } else {
                cond.write_verification_data( result->data_segment() );
            }
            
            result->on_cond_true( reinterpret_cast<ConditionedTreeNode *>( __sim_alloc( __cond_data_segment_approx_size, CONDITION_TYPE::required_alignment() ) ) );
        
            return result;
        } 
        
#else
        template<class CONDITION_TYPE>
		inline AbstractIntermediateTreeNode * createChildNode ( AbstractIntermediateTreeNode * const parent,
                                                                RayReferenceContainer & candidates,
                                                                CONDITION_TYPE && cond,                            
                                                                const uint_fast16_t index ) {
            AbstractIntermediateTreeNode * result = NULL;

	        if ( mVisitor->state( index ) == TERMINATE || mVisitor->state( index ) == CHANGE_VISITOR ) {
	            if ( !candidates.empty() ) {
                    LeafNode<CONDITION_TYPE> * leaf = new LeafNode<CONDITION_TYPE>( candidates, std::move( cond ) );
		            parent->insert( leaf );
		        }
	        } else {        
                IntermediateTreeNode<CONDITION_TYPE> * intermediate = new IntermediateTreeNode<CONDITION_TYPE>( std::move( cond ), mVisitor->getSplitFactor() );
                parent->insert( intermediate );
                result = intermediate;
	        }   
            
            return result; 
        }
#endif /* __DD_RECURSIVE_QUERY__ */
        
        template<class CONDITION_TYPE>
		inline ConditionedRayReferenceContainerList * createChildNode ( ConditionedRayReferenceContainerList * const parent,
                                                                        RayReferenceContainer & candidates,
                                                                        CONDITION_TYPE && cond,                            
                                                                        const uint_fast16_t index ) {
            ConditionedRayReferenceContainerList * result = NULL;

	        if ( mVisitor->state( index ) == TERMINATE || mVisitor->state( index ) == CHANGE_VISITOR ) {
	            if ( !candidates.empty() ) {
		            parent->insert( new ConditionedRayReferenceContainer<CONDITION_TYPE>( candidates, std::move( cond ) ) );
		        }
	        } else {
                result = parent;
                partition<CONDITION_TYPE, ConditionedRayReferenceContainerList>( parent, candidates );  
	        }    
            
            return result;
        }    
        

        
        template<class CONDITION_TYPE, class NODE_TYPE>
        void partition ( NODE_TYPE * const node,
                         RayReferenceContainer & candidates,
                         const bool terminate = false ) {

            const uint_fast16_t split_factor = mVisitor->getSplitFactor();
            
            RayReferenceContainer set[ 2 ]; // = new RayReferenceContainer[ split_factor ];
            CONDITION_TYPE * condition = new CONDITION_TYPE[ split_factor ];
            bool * leaf_flag = new bool[ split_factor ];

            mVisitor->registerOutputSets( set );
            mVisitor->before( candidates );

            candidates.successive_pop( *mVisitor );
            candidates.destroy();

            for ( uint_fast16_t index = 0; index < split_factor; index++ ) {
                mVisitor->refineCondition( index, condition[ index ] );
                leaf_flag[ index ] = ( mVisitor->state( index ) != CONTINUE );
            }  
            
            NODE_TYPE * child_node_1 = createChildNode( node, set[ 1 ], std::move( condition[ 1 ] ), leaf_flag[ 1 ] );
    #ifdef __OUTPUT_TREES__
            child_node_1->__parent = node;
            child_node_1->__label = condition[ 1 ].toString();
    #endif
            if ( !leaf_flag[ 1 ] ) {
                partition<CONDITION_TYPE, NODE_TYPE>( child_node_1, set[ 1 ] );
            } 
    #ifndef __DD_RECURSIVE_QUERY__
            child_node_1->on_cond_false( reinterpret_cast<NODE_TYPE *>( __sim_alloc( __cond_data_segment_approx_size, CONDITION_TYPE::required_alignment() ) ) );
    #endif
            
            NODE_TYPE * child_node = createChildNode( node, set[ 0 ], std::move( condition[ 0 ] ), leaf_flag[ 0 ] );
    #ifdef __OUTPUT_TREES__
                child_node->__parent = node;
                child_node->__label = condition[ 0 ].toString();
    #endif
            if ( !leaf_flag[ 0 ] ) {
                partition<CONDITION_TYPE, NODE_TYPE>( child_node, set[ 0 ], terminate );
    #ifndef __DD_RECURSIVE_QUERY__
                if ( terminate ) {
                    child_node->on_cond_false( NULL );
                } else {
                    child_node->on_cond_false( reinterpret_cast<NODE_TYPE *>( __sim_alloc( __cond_data_segment_approx_size, CONDITION_TYPE::required_alignment() ) ) );
                }
    #endif
            } else {
    #ifndef __DD_RECURSIVE_QUERY__
                if ( terminate ) {
                    child_node->on_every_cond( NULL );
                } else {
                    child_node->on_cond_false( reinterpret_cast<NODE_TYPE *>( __sim_alloc( __cond_data_segment_approx_size, CONDITION_TYPE::required_alignment() ) ) );
                }
    #else
                /* NOOP */;
    #endif
            }
                          
            delete[] condition;
            delete[] leaf_flag;

            mVisitor->after();

        //    delete[] set;
        }    
        
        template<class CONDITION_TYPE>
        const TreeNode * createTree ( RayReferenceContainer & candidates ) {

            if ( candidates.size() < mMinElements || mVisitor->leaf_condition( candidates ) ) {
                return new AbstractLeafNode( candidates );
            } else {
                        
                mVisitor->init( candidates );        
#ifdef __DD_RECURSIVE_QUERY__
                RootTreeNode<TRANSFORMATOR> * root;
                root = new RootTreeNode<TRANSFORMATOR>( mVisitor->getSplitFactor(), mTransformator );
                partition<CONDITION_TYPE, AbstractIntermediateTreeNode>( root, candidates ); 
#else
                RootTreeNode<TRANSFORMATOR, CONDITION_TYPE> * root;
                start_new_memory_region( sizeof( ConditionedTreeNode ) + CONDITION_TYPE::dynamic_size(), candidates.size() );
                __cond_data_segment_approx_size = sizeof( ConditionedTreeNode ) + CONDITION_TYPE::guaranteed_alignment() - sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
                
                char * __root_mem = __alloc( sizeof( RootTreeNode<TRANSFORMATOR, CONDITION_TYPE> ) );
                root = new( __root_mem ) RootTreeNode<TRANSFORMATOR, CONDITION_TYPE>( mTransformator );
                root->set_entry_point( reinterpret_cast<ConditionedTreeNode *>( __sim_alloc( __cond_data_segment_approx_size, CONDITION_TYPE::required_alignment() ) ) );
                
                ConditionedTreeNode fake_root;
                partition<CONDITION_TYPE, ConditionedTreeNode>( &fake_root, candidates, true );
#endif
                
                return root;
            }
        }        
        
        inline void start_new_memory_region ( const uint_fast16_t node_size,
                                              const uint_fast32_t candidates ) {
            uint_fast32_t max_nodes = 1;
                
            for ( uint_fast16_t level = 1; level <= mVisitor->max_depth(); ++level ) {
                max_nodes += pow( 2 /*mVisitor->getSplitFactor()*/, level ); // @FIXME Put some estimation here!
            }
            
            char ** memory_regions = new char*[ __memory_regions_count + 1 ];
            uintptr_t ** ref_array_regions = new uintptr_t*[ __memory_regions_count + 1 ];
            
            for ( uint_fast16_t reg_index = 0; reg_index < __memory_regions_count; ++reg_index ) {
                memory_regions[ reg_index ] = __memory_regions[ reg_index ];
                ref_array_regions[ reg_index ] = __ref_array_regions[ reg_index ];
            }
            memory_regions[ __memory_regions_count ] = static_cast<char*>( aligned_alloc( sizeof( __BIT_LIST_RAW_DATA_TYPE__ ), max_nodes * node_size +  candidates * __DD_BLOCK_COUNT__ * sizeof( __BIT_LIST_RAW_DATA_TYPE__ ) )); 
            ref_array_regions[ __memory_regions_count ] = new uintptr_t[ candidates ];
            
            if ( __memory_regions ) {
                delete[] __memory_regions;
            }
            
            if ( __ref_array_regions ) {
                delete[] __ref_array_regions;
            }
            
             __memory_regions = memory_regions;
             __ref_array_regions = ref_array_regions;
             
             __memory = __memory_regions[ __memory_regions_count ];
             __ref_array = __ref_array_regions[ __memory_regions_count++ ];
        }
        
        public:
        TreeCreator ( SPLITTER * const visitor,
                      const TRANSFORMATOR * const trans,
                      const uint_fast32_t min_elements ) :
    #ifndef __DD_RECURSIVE_QUERY__
            __memory_regions( NULL ),
            __ref_array_regions( NULL ),
            __memory_regions_count( 0 ),
    #endif
            mVisitor( visitor ),
            mTransformator( trans ),
            mMinElements( min_elements ) {}
            
        ~TreeCreator () {
            delete mTransformator;
            delete mVisitor;
    #ifndef __DD_RECURSIVE_QUERY__
            for ( uint_fast16_t reg_index = 0; reg_index < __memory_regions_count; ++reg_index ) {
                free( __memory_regions[ reg_index ] );
                delete[] __ref_array_regions[ reg_index ];
            }
            if ( __memory_regions ) {
                delete[] __memory_regions;
                delete[] __ref_array_regions;
            }
    #endif
        }
        
        const TreeNode * createHDTree ( RayReferenceContainer && candidates ) {
            return createTree<HD_COND_TYPE>( candidates );
        }
        
        const TreeNode * createLDTree ( RayReferenceContainer && candidates ) {
            return createTree<LD_COND_TYPE>( candidates );
        }        

#ifdef __DD_RECURSIVE_QUERY__        
        const ConditionedRayReferenceContainerList * createList ( RayReferenceContainer && candidates ) {
            ConditionedRayReferenceContainerList * const list = new ConditionedRayReferenceContainerList();
            
            mVisitor->init( candidates );
            partition<HD_COND_TYPE, ConditionedRayReferenceContainerList>( list, candidates );
            
            return list;
        }
#else
        const ConditionedRayReferenceContainerList * createList ( RayReferenceContainer && ) {
            return NULL;
        }
#endif
        
        inline uint_fast16_t getAdjacencyCandidateSize () const {
            return mTransformator->getAdjacencyCandidateSize();
        }
        
        inline uint_fast16_t getAdjacencyStackSize () const {
            return mTransformator->getAdjacencyStackSize();
        }
    };
}

#endif /* __DD_TREE_CREATOR__ */
