/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_TREE_NODE_CONDITION__
#define __DD_TREE_NODE_CONDITION__

#include "DDBitList.h"
#include "DDConstants.h"

#include "DDRayReference.h"
#include "DDRayReferenceContainer.h"
#include "DDEdgeCandidatesContainer.h"
#include "DDRayReferenceOverapproximation.h"

namespace addibit {

	class TreeNodeCondition {

        protected:
        BitList mMap;        
        
        inline BitList & union_map () {
            return mMap;
        }

		public:
#ifdef __STATS_OUTPUT__
        static uint64_t NARROWING_OPS;
        static uint64_t VERIFICATION_OPS;
        static uint64_t NONTRIVIAL_CANDIDATES;

        static void start_new_stat_session () {
            NARROWING_OPS = 0;
            VERIFICATION_OPS = 0;
            NONTRIVIAL_CANDIDATES = 0;
        }
#endif

        TreeNodeCondition () {}            
        
        TreeNodeCondition ( const TreeNodeCondition & lvalue ) :
            mMap( lvalue.mMap ) {}
    
        TreeNodeCondition ( BitList && map ) :
            mMap( std::move( map ) ) {}
            
        TreeNodeCondition ( TreeNodeCondition && rvalue ) :
            mMap( std::move( rvalue.mMap ) ) {} 

		virtual ~TreeNodeCondition() {}

        inline void operator= ( TreeNodeCondition && rvalue ) {
            mMap = std::move( rvalue.mMap );
        }

        inline void operator= ( const TreeNodeCondition & lvalue ) {
            mMap = lvalue.mMap;
        }

        inline const BitList & union_map () const {
            return mMap;
        }

        virtual bool acceptsContainer () const {
            return false;
        }

        static uint_fast16_t dynamic_size () {
            return __DD_BLOCK_COUNT__ * sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
        
        static uint_fast16_t required_alignment () {
            return sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
        
        static uint_fast16_t guaranteed_alignment () {
            return sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }

        uint_fast16_t mem_size () const {
            return __DD_BLOCK_COUNT__ * sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }

        inline bool proceed_narrowing( const NarrowingCandidate & candidate ) const {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::NARROWING_OPS++;
    #endif 
            return mMap.getCommonPopCnt( candidate.data() ) >= GLOBAL_EDGE_DIM;
        }
        
        void write_data ( __BIT_LIST_RAW_DATA_TYPE__ * const data_region ) const {
            
            data_region[ 0 ] = mMap[ 0 ];
            for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__; ++block ) {
                data_region[ block ] = mMap[ block ];
            }
        }

        inline void write_narrowing_data ( __BIT_LIST_RAW_DATA_TYPE__ * const data_region ) const {
            write_data( data_region );
        }
        
        inline void write_verification_data ( __BIT_LIST_RAW_DATA_TYPE__ * const data_region ) const {
            write_data( data_region );
        }
        
        inline static bool narrow_feasible ( EdgeCandidatesContainer & container, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data  ) {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::NARROWING_OPS++;
    #endif
            const NarrowingCandidate * const candidate = container.candidate();
            
            return BitList::getCommonPopCnt( candidate->data(), node_data ) >= GLOBAL_EDGE_DIM;
        }         
        
        inline static void intermediate_post_action ( EdgeCandidatesContainer & , const __BIT_LIST_RAW_DATA_TYPE__ * const ) {}
        inline static void leaf_post_action ( EdgeCandidatesContainer & , const __BIT_LIST_RAW_DATA_TYPE__ * const ) {}
        inline static bool intermediate_post_action ( EdgeCandidate &, const __BIT_LIST_RAW_DATA_TYPE__ * const ) {
            return false;
        }
        inline static bool leaf_post_action ( EdgeCandidate &, const __BIT_LIST_RAW_DATA_TYPE__ * const ) {
            return false;
        }
        
        inline static bool verify_feasible ( const EdgeCandidate & candidate, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::VERIFICATION_OPS++;
    #endif 
			 return BitList::a_contains_b( node_data, candidate.data() );
        }

        inline bool proceed_verification ( const EdgeCandidate & candidate ) const {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::VERIFICATION_OPS++;
    #endif
            return mMap.contains( candidate.data() ); 
        }

        bool feasible ( const BitList & set_1 , const BitList & set_2 ) const {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::NARROWING_OPS++;
    #endif
            return set_1.getCommonPopCnt( set_2 ) >= GLOBAL_EDGE_DIM;
        }

        inline bool feasible ( const BitList & set_1 , const unsigned long * set_2 ) const {

    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::NARROWING_OPS++;
    #endif
            return set_1.getCommonPopCnt( set_2 ) >= GLOBAL_EDGE_DIM;
        }

        virtual const std::string toString () const;
	};
}

#endif // __DD_TREE_NODE_CONDITION__
