/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_TWO_BITS_FRAGMENTATION_CONDITION__
#define __DD_TWO_BITS_FRAGMENTATION_CONDITION__

#include "DDTreeNodeCondition.h"

#define __FRAGMENT_SIZE__   2

namespace addibit {

    class TwoBitsFragmentationCondition : public TreeNodeCondition {
    
        private:
        BitList mReductionMap;
        
        const BitList createReductionMap ( const unsigned int, const BitList * const &, const Interval * const & ) const;
    
        public:        
        TwoBitsFragmentationCondition () {}
                
        TwoBitsFragmentationCondition ( BitList && union_map,
                                        BitList && reduction_map ) :
            TreeNodeCondition( std::move( union_map ) ),
            mReductionMap( std::move( reduction_map ) ) {}
                
        TwoBitsFragmentationCondition ( TwoBitsFragmentationCondition && rvalue ) :
            TreeNodeCondition( std::move( rvalue ) ),
            mReductionMap( rvalue.mReductionMap ) {}
            
        void init ( const BitList & reduction ) {
            mReductionMap = reduction;
            union_map() &= ~mReductionMap;
        }              

        void init ( const uint_fast16_t fcount,
                    const BitList * const & fragments,
                    const Interval * const & bounds ) {
                    
            mReductionMap = createReductionMap( fcount, fragments, bounds );
            union_map() &= ~mReductionMap;
        }           
        
        inline void operator= ( const TwoBitsFragmentationCondition & lvalue ) {
            TreeNodeCondition::operator= ( lvalue );
            mReductionMap = lvalue.mReductionMap;
        }
        
        void write_data ( __BIT_LIST_RAW_DATA_TYPE__ * const data_region ) const {
            
            data_region[0] = mMap[ 0 ];
            data_region[ __DD_BLOCK_COUNT__ ] = mReductionMap[ 0 ];

            for ( uint_fast16_t block = 1; block < __DD_BLOCK_COUNT__; ++block ) {
                data_region[ block ] = mMap[ block ];
                data_region[ block + __DD_BLOCK_COUNT__ ] = mReductionMap[ block ];
            }
        }
        
        inline void write_narrowing_data ( __BIT_LIST_RAW_DATA_TYPE__ * data_region ) const {
            write_data( data_region );
        }

        inline void write_verification_data ( __BIT_LIST_RAW_DATA_TYPE__ * data_region ) const {
            write_data( data_region );
        }
        
        inline static bool narrow_feasible ( EdgeCandidatesContainer & container, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::NARROWING_OPS++;
    #endif
            const NarrowingCandidate * const candidate = container.candidate();
            
            return BitList::getReducedCommonPopCnt( node_data, candidate->data(), node_data + __DD_BLOCK_COUNT__ ) >= GLOBAL_EDGE_DIM;
        }
        
        inline static void intermediate_post_action ( EdgeCandidatesContainer & , const __BIT_LIST_RAW_DATA_TYPE__ * const ) {}
        inline static void leaf_post_action ( EdgeCandidatesContainer & , const __BIT_LIST_RAW_DATA_TYPE__ * const ) {}
        inline static bool intermediate_post_action ( EdgeCandidate &, const __BIT_LIST_RAW_DATA_TYPE__ * const ) {
            return false;
        }
        inline static bool leaf_post_action ( EdgeCandidate &, const __BIT_LIST_RAW_DATA_TYPE__ * const ) {
            return false;
        }
        
        inline static bool verify_feasible ( const EdgeCandidate & candidate, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::VERIFICATION_OPS++;
    #endif 
			 return BitList::reduced_a_contains_b( node_data, candidate.data(), node_data + __DD_BLOCK_COUNT__ ); 
        }
        
        bool proceed_narrowing( const NarrowingCandidate & candidate ) const {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::NARROWING_OPS++;
    #endif
            return union_map().getReducedCommonPopCnt( candidate.data(), mReductionMap ) >= GLOBAL_EDGE_DIM;
        }   
        
        bool proceed_verification ( const EdgeCandidate & candidate ) const {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::VERIFICATION_OPS++;
    #endif
    
            return union_map().reduced_a_contains_b( candidate.data(), mReductionMap );
        }    
        
        static uint_fast16_t dynamic_size () {
            return TreeNodeCondition::dynamic_size() + __DD_BLOCK_COUNT__ * sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
        
        static uint_fast16_t required_alignment () {
            return sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
        
        static uint_fast16_t guaranteed_alignment () {
            return sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
    };
}

#endif /* __DD_TWO_BITS_FRAGMENTATION_CONDITION__ */
