/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_VANTAGE_POINT_PRELIMINARY_CONDITION__
#define __DD_VANTAGE_POINT_PRELIMINARY_CONDITION__

#include "DDInterval.h"

namespace addibit {
    
    union VantagePointNodeData {
        __m64 __data;
        int32_t __vec[2];
    };
    
    union VantagePointCmpData {
        __m64 __data;
        int64_t __cmp_value;        
    };

	class VantagePointTreeNodeCondition {

        private:

        void write_data ( __BIT_LIST_RAW_DATA_TYPE__ * data_region ) const {
                        
            *( data_region++ ) = __vp_id;
            
            uint_fast16_t i = 0;
            
            do {
                data_region[ i ] = mVantagePoint[ i ];
            } while ( ++i < __DD_BLOCK_COUNT__ );      
        }
        
        protected:
        BitList mVantagePoint;
        uintptr_t __vp_id;

        Interval mDistanceRange;
        Interval mInverseRange;

		public:
        VantagePointTreeNodeCondition () {}
        
        VantagePointTreeNodeCondition ( const VantagePointTreeNodeCondition & lvalue ) :
            mVantagePoint( lvalue.mVantagePoint ),
            __vp_id( lvalue.__vp_id ),
            mDistanceRange( lvalue.mDistanceRange ),
            mInverseRange( lvalue.mInverseRange ) {}        
        
        VantagePointTreeNodeCondition ( VantagePointTreeNodeCondition && rvalue ) :
            mVantagePoint( std::move( rvalue.mVantagePoint ) ),
            __vp_id( rvalue.__vp_id ),
            mDistanceRange( rvalue.mDistanceRange ),
            mInverseRange( rvalue.mInverseRange ) {}

		VantagePointTreeNodeCondition ( BitList && v_point,
                                        const Interval & distance,
                                        const Interval & inverse,
                                        const uintptr_t vp_id ) :
            mVantagePoint( std::move( v_point ) ),
            __vp_id( vp_id ),
            mDistanceRange( distance ),
            mInverseRange( inverse ) {}

        inline void operator= ( VantagePointTreeNodeCondition && rvalue ) {
            mVantagePoint = std::move( rvalue.mVantagePoint );
            mDistanceRange = rvalue.mDistanceRange;
            mInverseRange = rvalue.mInverseRange;
            __vp_id = rvalue.__vp_id;
        }
        
        inline void operator= ( const VantagePointTreeNodeCondition & lvalue ) {
            mVantagePoint = lvalue.mVantagePoint;
            mDistanceRange = lvalue.mDistanceRange;
            mInverseRange = lvalue.mInverseRange;
            __vp_id = lvalue.__vp_id;
        }
        
        void write_narrowing_data ( __BIT_LIST_RAW_DATA_TYPE__ * data_region ) const {
            
            VantagePointNodeData * range = reinterpret_cast<VantagePointNodeData *>( data_region++ );
            range->__vec[0] = (int32_t)mInverseRange.ub() - (int32_t)GLOBAL_EDGE_DIM;
            range->__vec[1] = (int32_t)mDistanceRange.ub() - (int32_t)GLOBAL_EDGE_DIM;
            
            write_data( data_region );
        }
        
        void write_verification_data ( __BIT_LIST_RAW_DATA_TYPE__ * data_region ) const {
            
            VantagePointNodeData * range = reinterpret_cast<VantagePointNodeData *>( data_region++ );
            range->__vec[0] = (int32_t)mDistanceRange.ub();
            range->__vec[1] = (int32_t)mInverseRange.ub();
            
            write_data( data_region );
        }        

        inline static bool narrow_feasible ( EdgeCandidatesContainer & container, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::NARROWING_OPS++;
    #endif
            AdjacencyTestStack * const stack = container.stack();

            const VantagePointNodeData * node_range = reinterpret_cast<const VantagePointNodeData *>( node_data );
            const VantagePointNodeData * cand_range = reinterpret_cast<const VantagePointNodeData *>( stack->pop() );
            
            VantagePointNodeData res;
            res.__data = _mm_add_pi32( node_range->__data, cand_range->__data );
            
            return res.__vec[0] >= 0 && res.__vec[1] >= 0;
        }         
        
        inline static void intermediate_post_action ( EdgeCandidatesContainer & container, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
            const NarrowingCandidate * const candidate = container.candidate();
            AdjacencyTestStack * const stack = container.stack();

    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::NARROWING_OPS++;
    #endif
            
            VantagePointNodeData * new_aux_data = reinterpret_cast<VantagePointNodeData *>( stack->push( 2 ) );
            new_aux_data->__vec[0] = BitList::getCommonPopCnt( node_data + 2, candidate->data() );
            new_aux_data->__vec[1] = candidate->population() - new_aux_data->__vec[0];
            if ( new_aux_data->__vec[0] >= GLOBAL_EDGE_DIM ) {
               container.push( node_data + 2, node_data[1] ); 
            }
            new_aux_data[1] = new_aux_data[0];
        }
        
        inline static void leaf_post_action ( EdgeCandidatesContainer & container, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
            const NarrowingCandidate * const candidate = container.candidate();
            
            if ( BitList::getCommonPopCnt( node_data + 2, candidate->data() ) >= GLOBAL_EDGE_DIM ) {
                container.push( node_data + 2, node_data[1] ); 
            }
        }
        
        inline static bool intermediate_post_action ( EdgeCandidate & candidate, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
            const AdjacencyCandidate & adj_cand = candidate.adj_cand();
            AdjacencyTestStack * const stack = candidate.stack();
            
            VantagePointNodeData * new_aux_data = reinterpret_cast<VantagePointNodeData *>( stack->push( 2 ) );
            new_aux_data->__vec[0] = BitList::getCommonPopCnt( node_data + 2, adj_cand.data() );
            new_aux_data->__vec[1] = candidate.population() - new_aux_data->__vec[0];
            new_aux_data[1] = new_aux_data[0];
            
            return new_aux_data->__vec[1] == 0 && candidate.illegal_1() != node_data[1] && candidate.illegal_2() != node_data[1]; 
        }
        
        inline static bool leaf_post_action ( EdgeCandidate & candidate, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
            return  BitList::a_contains_b( node_data + 2, candidate.adj_cand().data() ) && candidate.illegal_1() != node_data[1] && candidate.illegal_2() != node_data[1];
        }
        
        inline static bool verify_feasible ( EdgeCandidate & candidate, const __BIT_LIST_RAW_DATA_TYPE__ * const node_data ) {
    #ifdef __STATS_OUTPUT__
            TreeNodeCondition::VERIFICATION_OPS++;
    #endif 

            const VantagePointNodeData * node_range = reinterpret_cast<const VantagePointNodeData *>( node_data );
            const VantagePointNodeData * cand_range = reinterpret_cast<const VantagePointNodeData *>( candidate.stack()->pop() );
            
            VantagePointCmpData __res;
            __res.__data = _mm_cmpgt_pi32( cand_range->__data, node_range->__data );
            
            return __res.__cmp_value == 0; 
        }

        const std::string toString () const {
            std::stringstream result;

            result << "dist=" << mDistanceRange.toString() << ", "
                   << "inv="  << mInverseRange.toString() << ", "
                   << "|v|="  << mVantagePoint.getPopulationCount();

            return result.str();            
        }
        
        uint_fast16_t mem_size () const {
             return ( __DD_BLOCK_COUNT__ + 2 ) * sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
        
        static uint_fast16_t dynamic_size () {
            return (  __DD_BLOCK_COUNT__ + 2 ) * sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
        
        static uint_fast16_t required_alignment () {
            return sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
        
        static uint_fast16_t guaranteed_alignment () {
            return sizeof( __BIT_LIST_RAW_DATA_TYPE__ );
        }
	};
}

#endif // __DD_VANTAGE_POINT_PRELIMINARY_CONDITION__
