/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_VANTAGE_POINT_VISITOR__
#define __DD_VANTAGE_POINT_VISITOR__

#include <algorithm>

#include "DDTweaks.h"

#include "DDBitSplitVisitor.h"
#include "DDContainerVantagePointMapping.h"
#include "DDVantagePointTreeNodeCondition.h"

#define BINARY_VP_SUBSET        0
#define BINARY_VP_SUPERSET      1

namespace addibit {
    
    class VantagePointSort {
        
        private:
        const uint_fast16_t * __ref_dist;
      
        public:
        VantagePointSort( const uint_fast16_t * ref_dist ) :
           __ref_dist( ref_dist ) {}
           
        bool operator() ( const RayReference & first, const RayReference & second ) const {
            return first.degeneracy > second.degeneracy or ( first.degeneracy == second.degeneracy and __ref_dist[ first.id ] > __ref_dist[ second.id ] );
        }
    };

    class VantagePointVisitor : public BitSplitVisitor {

        protected:
        uint_fast16_t mMaxDistance;
        const double_t mDistanceCoeff;
        ContainerVantagePointMapping mVantagePointMapping;
        
        const BitList & __static_vp;

        void registerContainer ( const unsigned int ) {
            mVantagePointMapping.insert_empty();
        }
        
        unsigned int getTargetContainer ( const RayReference & ) {
            return 0;
        }

        public:
        VantagePointVisitor ( const uint_fast16_t depth,
                              const uint_fast16_t capacity,
                              const BitList & static_vp,
                              const double_t distance_coeff ) :
            BitSplitVisitor( depth, capacity, 2 ),
            mDistanceCoeff( distance_coeff ),
            mVantagePointMapping( depth, 2 ),
            __static_vp( static_vp ) {}

        void init ( RayReferenceContainer & container ) {
            BitSplitVisitor::init( container );
            
            mVantagePointMapping.init( container.size() );
            mVantagePointMapping[0].set_vantage_point( RayReference( __static_vp ) );
            container.iterate( mVantagePointMapping );
            
            std::sort( container.bucket(0).start_ptr(), container.bucket(0).end_ptr(), VantagePointSort( mVantagePointMapping.ref_dist() ) );
        } 
        
        void before ( const RayReferenceContainer & container ) {
            BitSplitVisitor::before( container );
            const ContainerVantagePointEntry & entry = mVantagePointMapping[ container.id() ];

            mMaxDistance = entry.get_distance( container.size() * mDistanceCoeff );
        }
        
        void onPop ( RayReference && ref ) {
            
            const uint_fast16_t target = ( mVantagePointMapping.last_ref_dist( ref.id ) <= mMaxDistance ) ? 0 : 1;
            mVantagePointMapping.update( set( target ).id(), std::move( ref ) );  
            
            if ( likely( ref.activeSet.data() != NULL )) {
                set( target ).insert( std::move( ref ) );
            }
        }
        
        bool can_proceed ( const unsigned int index ) const {
            const uint_fast16_t cid = set( index ).id();
            return BitSplitVisitor::can_proceed( index ) && !mVantagePointMapping[ cid ].distance_range().pointed();
        }        
        
        void refineCondition ( const uint_fast32_t index,
                               VantagePointTreeNodeCondition & cond ) const {
            
            const uint_fast32_t container_id = set( index ).id();
            const ContainerVantagePointEntry & entry = mVantagePointMapping[ container_id ];
            
            if ( likely( entry.vantage_point().activeSet.data() != NULL ) ) {
                cond = VantagePointTreeNodeCondition( BitList( entry.vantage_point().activeSet ), mVantagePointMapping.conditional_dist( container_id ), mVantagePointMapping.conditional_inv( container_id ), entry.vantage_point().__address );
            } else {
                cond = VantagePointTreeNodeCondition( BitList( __BIT_LIST_ALLOC_SPACE__, 0 ), Interval(0,0), Interval(0,0), 0x00000000 );
            }
            
            DD_LOG_DEBUG( "mDistanceRange = " << entry.distance_range().toString() << ", mPopCntRange = " << entry.inverse_range().toString() )
        }
    };
}

#endif // __DD_VANTAGE_POINT_VISITOR__
